import contextlib
import warnings


@contextlib.contextmanager
def ignore_widget_on_submit_warning():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=r".*on_submit is deprecated.*",
            category=DeprecationWarning,
        )
        yield



import pytest
from superintendent.distributed import ClassLabeller
import superintendent.base
import superintendent.distributed.queueing
import superintendent.controls


def test_that_class_labeller_creates_base_widget():

    widget = ClassLabeller()

    assert isinstance(widget, superintendent.base.Labeller)


def test_that_class_labeller_has_expected_attributes():

    widget = ClassLabeller(options=("a", "b"))

    assert isinstance(
        widget.queue, superintendent.distributed.queueing.DatabaseQueue
    )
    assert isinstance(widget.input_widget, superintendent.controls.Submitter)

    assert pytest.helpers.same_elements(
        widget.input_widget.options, ("a", "b")
    )



from numbers import Number
from typing import Callable, Dict, List, Optional, Text, Union

import ipywidgets as widgets
import traitlets


class ButtonGroup(widgets.HBox):
    """A group of buttons with output widgets underneath.

        Parameters
        ----------
        options : list
            A list of options for this button group.
        button_width : str, int, float, optional
            The width of each button as an HTML compatible string or a number.
            (the default is None, which leads to the width being divided
            between the buttons.)
        """

    options = traitlets.List(
        default_value=[], trait=traitlets.Unicode(), allow_none=True
    )
    submission_functions = traitlets.List(default_value=[], allow_none=True)
    button_width = traitlets.Union(
        [traitlets.Float(), traitlets.Integer(), traitlets.Unicode()],
        allow_none=True,
    )

    def __init__(
        self,
        options: List[str],
        button_width: Optional[Union[Number, Text]] = None,
        *args,
        **kwargs
    ):

        super().__init__(children=[], **kwargs)

        if button_width is None and len(options) > 0:
            self.button_width = max(1 / len(options), 0.1)
        else:
            self.button_width = button_width

        self.options = options

    @traitlets.observe("options")
    def rearrange_buttons(self, change):
        """Rearrange the buttons.

        Parameters
        ----------
        change : Any
            Any ol' change.

        """

        self.buttons = self.hints = {
            option: ButtonWithHint(option, self.button_width)
            for option in self.options
        }

        for button in self.buttons.values():
            button.on_click(self._handle_click)

        self.children = [self.buttons[option] for option in self.options]

    def on_click(self, func: Callable) -> None:
        """Add a function to the list of calls made after a click.

        Parameters
        ----------
        func : Callable
            The function to call when the button is clicked.
        """
        if not callable(func):
            raise ValueError(
                "You need to provide a callable object, but you provided "
                + str(func)
                + "."
            )
        self.submission_functions.append(func)

    def _handle_click(self, owner: widgets.Button) -> None:
        for func in self.submission_functions:
            func(owner)

    @traitlets.validate("button_width")
    def _valid_value(self, proposal: Dict[str, Union[float, int, Text]]):
        if isinstance(proposal["value"], float) and proposal["value"] <= 1:
            return "{}%".format(int(100 * proposal["value"]))
        elif isinstance(proposal["value"], (int, float)):
            return "{}px".format(int(proposal["value"]))
        elif isinstance(proposal["value"], str):
            return proposal["value"]
        else:  # pragma: no cover
            raise traitlets.TraitError(
                "Button_width can only be a float, an integer, or a string."
            )


class ButtonWithHint(widgets.VBox):

    description = traitlets.Unicode()

    def __init__(self, label: str, button_width: str, *args, **kwargs):
        """Create a button.

        Parameters
        ----------
        label : str
            The label for this button.
        button_width : str
            How wide you'd like this button to be.

        """

        kwargs["layout"] = kwargs.get(
            "layout", widgets.Layout(width=button_width)
        )

        super().__init__(children=[], *args, **kwargs)
        self.button = widgets.Button(
            description=str(label), layout=widgets.Layout(width="95%")
        )

        self.hint = widgets.Output()
        self.children = [self.button, self.hint]

        self.description = label
        widgets.link((self, "description"), (self.button, "description"))

    def on_click(self, func: Callable):
        """Add a function to the list of calls made after a click.

        Parameters
        ----------
        func : Callable
            The function to call when the button is clicked.
        """
        self.button.on_click(func)

    def __enter__(self):
        return self.hint.__enter__()

    def __exit__(self, *args, **kwargs):
        return self.hint.__exit__(*args, **kwargs)



# -*- coding: utf-8 -*-
"""Tools to supervise your clustering."""
from . import base
from .queueing import ClusterLabellingQueue


class ClusterSupervisor(base.Labeller):
    """
    A widget for labelling clusters.

    Parameters
    ----------
    features : np.ndarray, pd.Series. pd.DataFrame
        Your features.
    cluster_indices : np.ndarray, pd.Series
        The cluster label for each data point.
    representativeness : np.ndarray, pd.Series
        How representative of a cluster your data points are. This can be the
        probability of cluster membership (as in e.g. HDBSCAN), or cluster
        centrality (as in e.g. K-Means).
    ignore : tuple, list
        Which clusters should be ignored. By default, this is -1, as most
        clustering algorithms assign -1 to points not in any cluster.
    """

    def __init__(
        self,
        features,
        cluster_indices,
        representativeness=None,
        ignore=(-1,),
        **kwargs
    ):
        """Create a labelling widget."""

        super().__init__(features, **kwargs)

        self.queue = ClusterLabellingQueue(
            features, cluster_indices, representativeness
        )

        self._annotation_loop = self._annotation_iterator()
        next(self._annotation_loop)
        self._compose()

    def _annotation_iterator(self):
        """
        The method that iterates over the clusters and presents them for
        annotation.
        """

        self.progressbar.bar_style = ""

        for cluster_index, data in self.queue:

            self._display(data)
            sender = yield

            if sender["source"] == "__undo__":
                # unpop the current item:
                self.queue.undo()
                # unpop and unlabel the previous item:
                self.queue.undo()
                # try to remove any labels not in the assigned labels:
                self.input_widget.remove_options(
                    set(self.input_widget.options) - self.queue.list_labels()
                )
            elif sender["source"] == "__skip__":
                pass
            else:
                new_label = sender["value"]
                self.queue.submit(cluster_index, new_label)
                # self.input_widget.add_hint(new_label, datapoint)

            self.progressbar.value = self.queue.progress

        if self.event_manager is not None:
            self.event_manager.close()

        yield self._render_finished()

    @property
    def new_clusters(self):
        return self.queue.cluster_labels



"""Helper functions for displaying types of data."""
from typing import Callable, Dict, Union
from typing_extensions import Literal

import IPython.display
import ipywidgets as widgets

from .image import image_display_function


def default_display_function(feature):
    """
    A default function that displays the feature and adds some padding.

    Parameters
    ----------
    feature : np.ndarray, pd.Series, pd.DataFrame
        The feature(s) you want to display
    """
    # n_samples = min(n_samples, feature.shape[0])
    IPython.display.display(widgets.Box(layout=widgets.Layout(height="2.5%")))
    IPython.display.display(feature)
    IPython.display.display(widgets.Box(layout=widgets.Layout(height="2.5%")))


functions: Dict[str, Callable] = {
    "default": default_display_function,
    "image": image_display_function,
    "img": image_display_function,
}

Names = Union[Literal["default"], Literal["image"], Literal["img"]]



import ipywidgets as widgets
import traitlets
from collections import deque

from .._compatibility import ignore_widget_on_submit_warning
from .hintedmultiselect import HintedMultiselect
from .keycapture import DEFAULT_SHORTCUTS
from .submitter import Submitter
from .togglebuttongroup import ToggleButtonGroup


class MulticlassSubmitter(Submitter):
    def _on_key_down(self, event):
        if event["type"] == "keyup":
            pressed_option = self._key_option_mapping.get(event.get("key"))

            if pressed_option is not None:
                self._toggle_option(pressed_option)
            elif event.get("key") == "Enter":
                self._when_submitted({"source": "enter"})
            elif event.get("key") == "Backspace":
                self._when_submitted({"source": "backspace"})

    def _toggle_option(self, option):
        self.control_elements._toggle(option)

    def _submit(self, sender=None):

        value = self.control_elements.value

        for func in self.submission_functions:
            func(value)

        self.control_elements._reset()

    def _handle_new_option(self, sender: widgets.Text):
        new_option = sender.value
        if new_option in self.options:
            return
        self.options = self.options + [new_option]
        self._toggle_option(new_option)
        self._undo_queue.append(lambda: self.remove_options((new_option,)))

    def _undo(self, sender=None):
        if self._undo_queue:
            callback = self._undo_queue.pop()
            callback()
        else:
            super()._undo(sender=sender)

    @traitlets.observe("other_option", "options", "max_buttons")
    def _compose(self, change=None):

        self._undo_queue = deque([])

        self.options = [str(option) for option in self.options]
        self._key_option_mapping = {
            key: option for key, option in zip(DEFAULT_SHORTCUTS, self.options)
        }

        if len(self.options) <= self.max_buttons:
            # if we can display all options:
            self.control_elements = ToggleButtonGroup(self.options)
        else:
            self.control_elements = HintedMultiselect(self.options)

        self.submission_button = widgets.Button(
            description="Apply", button_style="success"
        )
        self.submission_button.on_click(self._submit)

        if self.other_option:
            self.other_widget = widgets.Text(
                value="",
                description="Other:",
                placeholder="Hit enter to submit.",
            )
            with ignore_widget_on_submit_warning():
                self.other_widget.on_submit(self._handle_new_option)
        else:
            self.other_widget = widgets.HBox([])

        self.children = [
            self.control_elements,
            widgets.HBox(
                [
                    self.other_widget,
                    widgets.HBox(
                        [
                            self.sort_button,
                            self.skip_button,
                            self.undo_button,
                            self.submission_button,
                        ]
                    ),
                ],
                layout=widgets.Layout(justify_content="space-between"),
            ),
        ]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# superintendent documentation build configuration file, created by
# sphinx-quickstart on Sun Feb 18 13:13:36 2018.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#

import pkg_resources


# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinx.ext.githubpages",
    "jupyter_sphinx.execute",
    "sphinx_automodapi.automodapi",
    "nbsphinx",
    "m2r",
    "numpydoc",
]

numpydoc_show_class_members = False
# autosummary_generate = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = [".rst", ".md"]
# source_parsers = {'.md': 'recommonmark.parser.CommonMarkParser'}
# source_suffix = '.rst'

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "superintendent"
copyright = "2018, Jan Freyberg"
author = "Jan Freyberg"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
# The short X.Y version.
version = pkg_resources.get_distribution("superintendent").version

# version_ns = {}
# version_file = pathlib.Path("..").resolve() / "superintendent" / "version.py"
# exec(version_file.read_text(), version_ns)
# version = version_ns["version"]

# The full version, including alpha/beta/rc tags.
release = version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "**.ipynb_checkpoints"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

html_css_files = [
    "css/hide-double-widgets.css",
]

html_favicon = "favicon.ico"
html_logo = "img/logo.png"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
html_sidebars = {
    "**": [
        "relations.html",  # needs 'show_related': True theme option to display
        "searchbox.html",
        "globaltoc.html",
    ]
}


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "superintendentdoc"


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, "superintendent", "superintendent Documentation", [author], 1)
]


# -- Options for nbsphinx -------------------------------------------------

nbsphinx_prolog = """
.. note::
    This page will display what superintendent widgets look like, but not
    respond to user input (as it's not connected to a backend).
"""

# if this is not set to empty string, widgets get displayed twice, see:
# https://nbsphinx.readthedocs.io/en/0.5.0/usage.html#nbsphinx_widgets_path
# https://github.com/spatialaudio/nbsphinx/issues/378
nbsphinx_widgets_path = ""


# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {"https://docs.python.org/": None}


def setup(app):
    app.add_stylesheet("css/custom.css")
    # app.add_directive('autoautosummary', AutoAutoSummary)



"""Interactive machine learning supervision."""
from .class_labeller import ClassLabeller
from .multiclass_labeller import MultiClassLabeller

__all__ = ["MultiClassLabeller", "ClassLabeller"]
__version__ = "0.5.0rc1"



import time

from superintendent.controls import Timer


def test_that_timer_stores_the_time():
    timer = Timer()
    with timer:
        time.sleep(0.01)
    assert timer._time >= 0.01


def test_that_timer_compares_correctly():
    timer = Timer()
    with timer:
        time.sleep(0.01)

    assert timer >= 0.01
    assert timer < 0.1
    assert timer == timer._time



import time
from math import nan
from contextlib import ContextDecorator


class Timer(ContextDecorator):
    """
    A timer object. Use as a context manager to time operations, and compare to
    numerical values (seconds) to run conditional code.

    Usage:

    .. code-block:: python

        from superintendent.controls import Timer
        timer = Timer()
        with timer:
            print('some quick computation')
        if timer < 1:
            print('quick computation took less than a second')

    """

    def __init__(self):
        self._time = nan
        self._t0 = nan

    def start(self):
        self._time = nan
        self._t0 = time.time()

    def stop(self):
        self._time = time.time() - self._t0

    def __enter__(self):
        self.start()

    def __exit__(self, *args):
        self.stop()

    def __eq__(self, other):
        return self._time == other

    def __lt__(self, other):
        return self._time < other

    def __le__(self, other):
        return self._time <= other

    def __gt__(self, other):
        return self._time > other

    def __ge__(self, other):
        return self._time >= other

    def __repr__(self):
        return "{} s".format(self._time)



"""Base class to inherit from."""

from collections import OrderedDict
from functools import partial
from typing import Any, Callable, Optional, Union

# import ipyevents
import IPython.display
import ipywidgets as widgets
import numpy as np
import sklearn.model_selection
import traitlets
from sklearn.base import BaseEstimator

from . import acquisition_functions, controls, display
from .queueing import BaseLabellingQueue, SimpleLabellingQueue


class Labeller(traitlets.HasTraits):
    """
    Data point labelling.

    This is a base class for data point labelling.
    """

    options = traitlets.List(list(), allow_none=True)

    def __init__(
        self,
        *,
        features: Optional[Any] = None,
        labels: Optional[Any] = None,
        queue: BaseLabellingQueue = SimpleLabellingQueue(),
        input_widget: Optional[widgets.Widget] = None,
        display_func: Union[display.Names, Callable] = "default",
        model: Optional[BaseEstimator] = None,
        eval_method: Optional[Callable] = None,
        acquisition_function: Optional[Callable] = None,
        shuffle_prop: float = 0.1,
        model_preprocess: Optional[Callable] = None,
        display_preprocess: Optional[Callable] = None,
    ):
        """
        Make a class that allows you to label data points.


        Parameters
        ----------

        features : np.ndarray, pd.DataFrame, sequence
            This should be either a numpy array, a pandas dataframe, or any
            other sequence object (e.g. a list). You can also add data later.
        labels : np.array, pd.Series, sequence
            The labels for your data, if you have some already.
        queue : BaseLabellingQueue
            A queue object. The interface needs to follow the abstract class
            superintendent.queueing.BaseLabellingQueue. By default,
            SimpleLabellingQueue (an in-memory queue using python's deque)
        input_widget : Optional[widgets.Widget]
            An input widget. This needs to follow the interface of the class
            superintendent.controls.base.SubmissionWidgetMixin
        display_func : str, func, optional
            A function that accepts a single data point and displays it. Any
            return values are ignored. For convenience, you can supply the
            strings "image" and "default", which are defined in
            superintendent.display.
        model : sklearn.base.BaseEstimator
            An sklearn-interface compliant model (that implements `fit`,
            `predict`, `predict_proba` and `score`).
        eval_method : callable
            A function that accepts three arguments - model, x, and y - and
            returns the score of the model. If None,
            sklearn.model_selection.cross_val_score is used.
        acquisition_function : callable
            A function that re-orders data points during active learning. This
            can be a function that accepts a numpy array (class probabilities)
            or a string referring to a function from
            superintendent.acquisition_functions.
        shuffle_prop : float
            The proportion of data points that is shuffled when re-ordering
            during active learning. This is to avoid biasing too much towards
            the model predictions.
        model_preprocess : callable
            A function that accepts x and y data and returns x and y data. y
            can be None (in which it should return x, None) as this function is
            used on the un-labelled data too.
        display_preprocess : callable
            A function that accepts a single data point and pre-processes it.
            For example, if you have MNIST data as a vector of shape (784,),
            this function could re-shape the data to shape (28, 28)
        """

        # the widget elements
        self.layout = widgets.VBox([])
        self.feature_output = widgets.Output()
        self.feature_display = widgets.Box(
            (self.feature_output,),
            layout=widgets.Layout(
                justify_content="center",
                padding="2.5% 0",
                display="flex",
                width="100%",
            ),
        )

        if input_widget is None:
            raise ValueError("No input widget was provided.")

        self.input_widget = input_widget

        self.input_widget.on_submit(self._apply_annotation)
        self.input_widget.on_undo(self._undo)
        self.input_widget.on_skip(self._skip)

        self.options = self.input_widget.options
        traitlets.link((self, "options"), (self.input_widget, "options"))

        if queue is None:
            raise ValueError("No queue implementation was provided.")

        self.queue = queue

        if features is not None:
            self.queue.enqueue_many(features, labels)

        self.progressbar = widgets.FloatProgress(
            max=1, description="Progress:"
        )
        self.top_bar = widgets.HBox([self.progressbar])

        if isinstance(display_func, str):
            self._display_func = display.functions[display_func]
        else:
            self._display_func = display_func

        self.display_preprocess = display_preprocess

        self.display_timer = controls.Timer()
        self.model_fit_timer = controls.Timer()

        self.model = model
        self.eval_method = eval_method

        self.acquisition_function = acquisition_function
        if isinstance(acquisition_function, str):
            self.acquisition_function = acquisition_functions.functions[
                acquisition_function
            ]

        self.shuffle_prop = shuffle_prop
        self.model_preprocess = model_preprocess

        # if there is a model, we need the interface components for it
        if self.model is not None:
            self.retrain_button = widgets.Button(
                description="Retrain",
                disabled=False,
                button_style="",
                tooltip="Click me",
                icon="refresh",
            )
            self.retrain_button.on_click(self.retrain)
            self.model_performance = widgets.HTML(value="")
            self.top_bar.children = (
                widgets.HBox(
                    [*self.top_bar.children],
                    layout=widgets.Layout(width="50%"),
                ),
                widgets.HBox(
                    [self.retrain_button, self.model_performance],
                    layout=widgets.Layout(width="50%"),
                ),
            )

        # the annotation implementation:
        self._annotation_loop = self._annotation_iterator()
        next(self._annotation_loop)  # kick off the loop

    def _annotation_iterator(self):
        """The annotation loop."""

        self.progressbar.bar_style = ""
        self._compose()
        for id_, x in self.queue:

            self._display(x)
            y = yield

            if y is None:
                pass
            else:
                self.queue.submit(id_, y)

            self.progressbar.value = self.queue.progress

        yield self._render_finished()

    @classmethod
    def from_images(cls, *, canvas_size=(500, 500), **kwargs):
        """Generate a labelling widget from images.

        This is a convenience function that creates a widget with an image
        display function. All other arguments for creating this widget need to
        be passed.

        Parameters
        ----------
        canvas_size : tuple
            The width / height that the image will be fit into. By default 500
            by 500.

        """
        display_func = partial(
            display.image_display_function, fit_into=canvas_size
        )

        kwargs["display_func"] = kwargs.get("display_func", display_func)

        return cls(**kwargs)

    def _apply_annotation(self, y):
        self._annotation_loop.send(y)

    def _undo(self):
        # unpop the current item:
        self.queue.undo()
        # unpop and unlabel the previous item:
        self.queue.undo()
        # try to remove any labels not in the assigned labels:
        if hasattr(self.input_widget, "remove_options"):
            self.input_widget.remove_options(
                set(self.input_widget.options) - self.queue.list_labels()
            )
        # send None into the iterator; returning to the loop
        self._annotation_loop.send(None)

    def _skip(self):
        self._annotation_loop.send(None)

    def add_features(self, features, labels=None):
        """
        Add data to the widget.

        This adds the data provided to the queue of data to be labelled. You
        Can optionally provide labels for each data point.

        Parameters
        ----------
        features : Any
            The data you'd like to add to the labelling widget.
        labels : Any, optional
            The labels for the data you're adding; if you have labels.
        """
        self.queue.enqueue_many(features, labels)
        # reset the iterator
        self._annotation_loop = self._annotation_iterator()
        self.queue.undo()
        next(self._annotation_loop)

    def _display(self, feature):
        if feature is not None:

            # if displaying takes longer than 0.5 seconds, we should render
            # a placeholder
            if self.display_timer > 0.5:
                self._render_processing()

            if self.display_preprocess is not None:
                feature = self.display_preprocess(feature)

            with self.display_timer:
                with self.feature_output:
                    IPython.display.clear_output(wait=True)
                    self._display_func(feature)
            self._compose()

    def _compose(self):
        self.layout.children = [
            self.top_bar,
            self.feature_display,
            self.input_widget,
        ]

    def _render_processing(self, message="Rendering..."):
        message = (
            "<h1>{}".format(message)
            + '<i class="fa fa-spinner fa-spin"'
            + ' aria-hidden="true"></i>'
        )
        processing_display = widgets.HTML(value=message)
        with self.feature_output:
            IPython.display.clear_output()
            display.default_display_function(processing_display)

    def _render_finished(self):

        self.progressbar.bar_style = "success"

        finish_celebration = widgets.Box(
            (widgets.HTML(value="<h1>Finished labelling 🎉!"),),
            layout=widgets.Layout(
                justify_content="center",
                padding="2.5% 0",
                display="flex",
                width="100%",
            ),
        )
        self.layout.children = [self.progressbar, finish_celebration]

    @property
    def new_labels(self):
        _, _, labels = self.queue.list_all()
        return labels

    def _ipython_display_(self):
        IPython.display.display(self.layout)

    def retrain(self, button=None):
        """Re-train the classifier you passed when creating this widget.

        This calls the fit method of your class with the data that you've
        labelled. It will also score the classifier and display the
        performance.

        Parameters
        ----------
        button : widget.Widget, optional
            Optional & ignored; this is passed when invoked by a button.
        """
        if self.model is None:
            raise ValueError("No model to retrain.")

        if not self.model_fit_timer < 0.5:
            self._render_processing(message="Retraining... ")

        self.model_fit_timer.start()

        _, labelled_X, labelled_y = self.queue.list_completed()

        if len(labelled_y) < 10:
            self.model_performance.value = (
                "Score: Not enough labels to retrain."
            )
            return

        if self.model_preprocess is not None:
            labelled_X, labelled_y = self.model_preprocess(
                labelled_X, labelled_y
            )

        # first, fit the model
        self.model.fit(labelled_X, labelled_y)

        # now evaluate. by default, using cross validation. in sklearn this
        # clones the model, so it's OK to do after the model fit.
        if self.eval_method is not None:
            self.performance = np.mean(
                self.eval_method(self.model, labelled_X, labelled_y)
            )
        else:
            self.performance = np.mean(
                sklearn.model_selection.cross_val_score(
                    self.model,
                    labelled_X,
                    labelled_y,
                    cv=3,
                    error_score=np.nan,
                )
            )

        self.model_performance.value = "Score: {:.3f}".format(self.performance)

        if self.acquisition_function is not None:
            ids, unlabelled_X = self.queue.list_uncompleted()

            if self.model_preprocess is not None:
                unlabelled_X, _ = self.model_preprocess(unlabelled_X, None)

            reordering = list(
                self.acquisition_function(
                    self.model.predict_proba(unlabelled_X),
                    shuffle_prop=self.shuffle_prop,
                )
            )

            new_order = OrderedDict(
                [(id_, index) for id_, index in zip(ids, list(reordering))]
            )

            self.queue.reorder(new_order)

        self.model_fit_timer.stop()

        # undo the previously popped item and pop the next one
        self.queue.undo()
        self._skip()



pytest_plugins = ["helpers_namespace"]  # noqa

# fix matplotlib import errors on Mac OS:
import sys

if sys.platform == "darwin":  # noqa
    import matplotlib  # isort:skip

    matplotlib.use("PS")  # isort:skip

import warnings
import collections
import pytest

import numpy as np
import pandas as pd

from hypothesis.errors import HypothesisDeprecationWarning
from hypothesis import settings, HealthCheck

warnings.simplefilter("ignore", HypothesisDeprecationWarning)

settings.register_profile(
    "travis-ci", suppress_health_check=(HealthCheck.too_slow,), deadline=1500
)

settings.load_profile("travis-ci")


@pytest.helpers.register
def same_elements(a, b):
    """Test if two things have the same elements, in different orders."""
    return collections.Counter(a) == collections.Counter(b)


@pytest.helpers.register
def no_shared_members(a, b):
    return (set(a) & set(b)) == set()


@pytest.helpers.register
def exact_element_match(a, b):
    if isinstance(a, np.ndarray) and isinstance(b, np.ndarray):
        try:
            return ((a == b) | (np.isnan(a) & np.isnan(b))).all()
        except TypeError:
            return (a == b).all()
    elif isinstance(a, pd.DataFrame) and isinstance(b, pd.DataFrame):
        a = a.reset_index(drop=True)
        b = b.reset_index(drop=True)
        return (
            ((a == b) | (a.isnull() & b.isnull())).all().all()
            or a.empty
            or b.empty
        )
    else:
        return all(
            [
                a_ == b_ or (np.isnan(a_) and np.isnan(b_))
                for a_, b_ in zip(a, b)
            ]
        )


@pytest.helpers.register
def recursively_list_widget_children(parent):
    childless = [
        widget for widget in parent.children if not hasattr(widget, "children")
    ]
    parents = [
        widget for widget in parent.children if hasattr(widget, "children")
    ]
    for widget in parents:
        childless += recursively_list_widget_children(widget)

    return childless



import json
from typing import Any, Optional

import numpy as np
import pandas as pd


class DataEncoder(json.JSONEncoder):
    def default(self, obj):
        """
        Serialize numpy or pandas objects to json.

        Parameters
        ----------
        obj : Any
            The object to serialise.
        """
        if isinstance(obj, np.ndarray):
            return {"__type__": "__np.ndarray__", "__content__": obj.tolist()}
        elif isinstance(obj, pd.DataFrame):
            return {
                "__type__": "__pd.DataFrame__",
                "__content__": obj.to_dict(orient="split"),
            }
        elif isinstance(obj, pd.Series):
            return {
                "__type__": "__pd.Series__",
                "__content__": {
                    "dtype": str(obj.dtype),
                    "index": list(obj.index),
                    "data": obj.tolist(),
                    "name": obj.name,
                },
            }


def data_decoder(obj):
    """Deserialise an object.

    Parameters
    ----------
    obj : Any
        The object to serialise.
    """

    if "__type__" in obj:
        if obj["__type__"] == "__np.ndarray__":
            return np.array(obj["__content__"])
        elif obj["__type__"] == "__pd.DataFrame__":
            return pd.DataFrame(**obj["__content__"])
        elif obj["__type__"] == "__pd.Series__":
            return pd.Series(**obj["__content__"])
    return obj


def data_dumps(obj: Any) -> Optional[str]:
    """Serialise an object.

    Parameters
    ----------
    obj : Any
        The object to serialise.
    """
    if obj is None:
        return None
    return json.dumps(obj, cls=DataEncoder)


def data_loads(obj: Optional[str]) -> Any:
    """Serialise an object.

    Parameters
    ----------
    obj : str
        The string to deserialise.
    """
    if obj is None:
        return None
    return json.loads(obj, object_hook=data_decoder)



from typing import Callable, Any


class SubmissionWidgetMixin:
    def on_submit(self, callback: Callable[[Any], None]) -> None:
        """Register a callback for the user "submitting".

        Parameters
        ----------
        callback : callable( Any )
            A function that accepts a value (the label). Any return values are
            ignored.

        """
        raise NotImplementedError(
            "Object of type {classname} does not "
            "implement on_submit, and is therefore not a valid "
            "SubmissionWidget."
        )

    def on_undo(self, callback: Callable):
        """Register a callback for the user "undoing".

        Parameters
        ----------
        callback : callable()
            A function that accepts no arguments; any return values are
            ignored.
        """
        raise NotImplementedError(
            "Object of type {classname} does not "
            "implement on_undo, and is therefore not a valid "
            "SubmissionWidget."
        )

    def on_skip(self, callback: Callable):
        """Register a callback for the user "skipping".

        Parameters
        ----------
        callback : callable()
            A function that accepts no arguments; any return values are
            ignored.
        """
        raise NotImplementedError(
            "Object of type {classname} does not "
            "implement on_skip, and is therefore not a valid "
            "SubmissionWidget."
        )



import pytest
from hypothesis import strategies as st
from hypothesis import given, settings

from superintendent.controls.dropdownbutton import DropdownButton


@settings(deadline=None)
@given(options=st.lists(st.text(), unique=True))
def test_that_options_are_set(options):

    widget = DropdownButton(options)
    assert widget.options == options == list(widget.dropdown.options)

    widget = DropdownButton([])
    widget.options = options
    assert widget.options == options == list(widget.dropdown.options)


def test_that_on_click_adds_callables_to_execution_list(mocker):

    mock_callable = mocker.Mock()
    widget = DropdownButton(["a", "b"])

    widget.on_click(mock_callable)
    assert widget.submission_functions == [mock_callable]

    widget._handle_click("test")

    assert mock_callable.call_args == (("test",),)

    with pytest.raises(ValueError):
        widget.on_click("non-callable")


def test_that_button_description_is_updated():
    widget = DropdownButton(["a", "b"])
    widget.dropdown.value = "a"
    widget.dropdown.value = "b"
    assert widget.button.description == "b"


def test_that_enter_exit_for_output_get_called(mocker):
    mock_enter = mocker.patch("ipywidgets.Output.__enter__")
    mock_exit = mocker.patch("ipywidgets.Output.__exit__")

    widget = DropdownButton(["a", "b"])

    with widget.hints["a"]:
        pass

    assert mock_enter.call_count == 1
    assert mock_exit.call_count == 1



import traitlets

import pytest
from hypothesis import strategies as st
from hypothesis import given, settings

from superintendent.controls.togglebuttongroup import (
    ToggleButtonGroup,
    ToggleButtonWithHint,
)


@settings(deadline=None)
@given(options=st.lists(st.text(), unique=True))
def test_that_options_are_set(options):

    widget = ToggleButtonGroup(options)
    assert widget.options == options
    assert pytest.helpers.same_elements(widget.buttons.keys(), options)
    assert pytest.helpers.same_elements(
        [button.description for button in widget.buttons.values()], options
    )

    widget = ToggleButtonGroup([])
    widget.options = options
    assert widget.options == options
    assert pytest.helpers.same_elements(widget.buttons.keys(), options)
    assert pytest.helpers.same_elements(
        [button.description for button in widget.buttons.values()], options
    )


def test_that_values_are_read_correctly():

    widget = ToggleButtonGroup(["a", "b", "c"])
    assert widget.value == []

    widget.buttons["a"].value = True
    assert widget.value == ["a"]

    widget.buttons["b"].value = True
    assert pytest.helpers.same_elements(widget.value, ["a", "b"])


def test_that_toggle_changes_values():

    widget = ToggleButtonGroup(["a", "b", "c"])
    assert widget.value == []

    widget._toggle("a")
    assert widget.value == ["a"]

    widget._toggle("a")
    assert widget.value == []


def test_that_reset_sets_all_values_to_false():

    widget = ToggleButtonGroup(["a", "b", "c"])
    assert widget.value == []

    widget._toggle("a")
    widget._toggle("b")
    assert pytest.helpers.same_elements(widget.value, ["a", "b"])

    widget._reset()
    assert widget.value == []


def test_that_button_width_is_set_correctly():
    widget = ToggleButtonGroup(["a", "b"])
    assert widget.button_width == "50%"

    widget.button_width = 0.2
    assert widget.button_width == "20%"

    widget.button_width = 200
    assert widget.button_width == "200px"

    widget.button_width = "87%"
    assert widget.button_width == "87%"

    with pytest.raises(traitlets.TraitError):
        widget.button_width = {"test-set", "lol"}


def test_that_enter_exit_for_output_get_called(mocker):
    mock_enter = mocker.patch("ipywidgets.Output.__enter__")
    mock_exit = mocker.patch("ipywidgets.Output.__exit__")

    button = ToggleButtonWithHint("Hi", "50%")

    with button:
        pass

    assert mock_enter.call_count == 1
    assert mock_exit.call_count == 1



"""Input and timing control widgets."""

from typing import Any, Callable, Dict, List, Optional, Sequence

import ipywidgets as widgets
import traitlets

from .._compatibility import ignore_widget_on_submit_warning
from .base import SubmissionWidgetMixin
from .buttongroup import ButtonGroup
from .dropdownbutton import DropdownButton


class Submitter(SubmissionWidgetMixin, widgets.VBox):
    """
    A flexible data submission widget.

    Submitter allows you to specifiy options, which can be chosen either via
    buttons or a dropdown, and a text field for "other" values.

    Parameters
    ----------
    options : list, tuple, optional
        The data submission options.
    max_buttons : int
        The number buttons you want to display. If len(options) >
        max_buttons, the options will be displayed in a dropdown instead.
    other_option : bool, optional
        Whether the widget should contain a text box for users to type in
        a value not in options.
    hint_function : fun
        A function that will be passed the hint for each label, that displays
        some output that will be displayed under each label and can be
        considered a hint or more in-depth description of a label. During image
        labelling tasks, this might be a function that displays an example
        image.
    hints : dict
        A dictionary with each element of options as a key, and the data that
        gets passed to hint_function as input.
    update_hints : bool
        Whether to update hints as you go through - for options that don't
        have any hints yet.
    """

    other_option = traitlets.Bool(True)
    options = traitlets.List(list(), allow_none=True)
    max_buttons = traitlets.Integer(12)

    def __init__(
        self,
        options: Sequence[str] = (),
        max_buttons: int = 12,
        other_option: bool = True,
        hint_function: Optional[Callable] = None,
        hints: Optional[Dict[str, Any]] = None,
        update_hints: bool = True,
        # shortcuts=None,
    ):
        """
        Create a widget that will render submission options.

        Note that all parameters can also be changed through assignment after
        you create the widget.
        """
        super().__init__([])
        self.submission_functions: List[Callable[[Any], None]] = []
        self.skip_functions: List[Callable[[], None]] = []
        self.undo_functions: List[Callable[[], None]] = []

        self.hint_function = hint_function

        self.hints = dict() if hints is None else hints
        if self.hint_function is not None:
            for option, feature in self.hints.items():
                self.hints[option] = widgets.Output()
                with self.hints[option]:
                    self.hint_function(feature)

        self.sort_button = widgets.Button(
            description="Sort options", icon="sort"
        )
        self.sort_button.on_click(self._sort_options)

        self.skip_button = widgets.Button(
            description="Skip", icon="fast-forward"
        )
        self.skip_button.on_click(self._skip)
        self.undo_button = widgets.Button(description="Undo", icon="undo")
        self.undo_button.on_click(self._undo)
        self.options = [str(option) for option in options]
        self.fixed_options = [option for option in self.options]

        self.max_buttons = max_buttons

        self.other_option = other_option

        self._compose()

    def add_hint(self, value, hint):
        """Add a hint to the widget.

        Parameters
        ----------
        value : str
            The label for which this hint applies.
        hint : Any
            The data point to use for the hint.
        """
        if (
            self.hint_function is not None
            and self.hints is not None
            and value not in self.hints
        ):
            with self.control_elements.hints[value]:
                self.hint_function(hint)

    def remove_options(self, values):
        """Remove options from the widget.

        Parameters
        ----------
        values : Sequence[str]
            The options to remove.
        """

        self.options = [
            option
            for option in self.options
            if option not in values or option in self.fixed_options
        ]

    def on_submit(self, callback: Callable[[Any], None]):
        """
        Add a function to call when the user submits a value.

        Parameters
        ----------
        callback : callable
            The function to be called when the widget is submitted.
        """
        if not callable(callback):
            raise ValueError(
                "You need to provide a callable object, but you provided "
                + str(callback)
                + "."
            )
        self.submission_functions.append(callback)

    def _submit(self, sender):
        """The function that gets called by submitting an option.

        This is called by the button / text field elements and shouldn't be
        called directly.
        """
        # figure out if it's a button or the text field
        if isinstance(sender, widgets.Text):
            value = sender.value
        else:
            value = sender.description

        if value is not None and value not in self.options:
            self.options = self.options + [value]

        for callback in self.submission_functions:
            callback(value)

        self._compose()

    def on_undo(self, callback: Callable[[], None]):
        """Provide a function that will be called when the user presses "undo".

        Parameters
        ----------
        callback : Callable[[], None]
            The function to be called. Takes no arguments and returns nothing.
        """
        self.undo_functions.append(callback)

    def _undo(self, sender=None):
        for callback in self.undo_functions:
            callback()

    def on_skip(self, callback: Callable[[], None]):
        """Provide a function that will be called when the user presses "Skip".

        Parameters
        ----------
        callback : Callable[[], None]
            The function to be called. Takes no arguments and returns nothing.
        """
        self.skip_functions.append(callback)

    def _skip(self, sender=None):
        for callback in self.skip_functions:
            callback()

    def _sort_options(self, change=None):
        self.options = list(sorted(self.options))

    @traitlets.observe("other_option", "options", "max_buttons")
    def _compose(self, change=None):

        if len(self.options) <= self.max_buttons:
            self.control_elements = ButtonGroup(self.options)
        else:
            self.control_elements = DropdownButton(self.options)

        self.control_elements.on_click(self._submit)

        if self.other_option:
            self.other_widget = widgets.Text(
                value="",
                description="Other:",
                placeholder="Hit enter to submit.",
            )
            with ignore_widget_on_submit_warning():
                self.other_widget.on_submit(self._submit)
        else:
            self.other_widget = widgets.HBox([])

        self.children = [
            self.control_elements,
            widgets.HBox(
                [
                    self.other_widget,
                    widgets.HBox(
                        [self.sort_button, self.skip_button, self.undo_button]
                    ),
                ],
                layout=widgets.Layout(justify_content="space-between"),
            ),
        ]



from collections import defaultdict
from typing import Callable, DefaultDict, Sequence

import ipywidgets as widgets
import traitlets


class DropdownButton(widgets.VBox):

    options = traitlets.List(
        trait=traitlets.Unicode(), default_value=list(), allow_none=True
    )
    submission_functions = traitlets.List(
        default_value=list(), allow_none=True
    )

    def __init__(self, options: Sequence[str], *args, **kwargs):
        """Create a dropdown button.

        Parameters
        ----------
        options : Sequence[str]
            The options to display in the widget.
        """

        super().__init__(*args, **kwargs)

        self.options = options

        self.dropdown = widgets.Dropdown(
            options=[str(option) for option in self.options],
            description="Label:",
        )
        widgets.dlink((self, "options"), (self.dropdown, "options"))
        self.dropdown.observe(self._change_selection)

        self.button = widgets.Button(
            description="Submit.",
            tooltip="Submit label.",
            button_style="success",
        )
        self.button.on_click(self._handle_click)

        self.hints: DefaultDict[str, widgets.Output] = defaultdict(
            widgets.Output
        )

        self.children = [
            widgets.HBox([self.dropdown, self.button]),
            self.hints[self.dropdown.value],
        ]

    def on_click(self, func: Callable) -> None:
        """Add a function to the list of calls made after a click.

        Parameters
        ----------
        func : Callable
            The function to call when the button is clicked.
        """
        if not callable(func):
            raise ValueError(
                "You need to provide a callable object, but you provided "
                + str(func)
                + "."
            )
        self.submission_functions.append(func)

    def _handle_click(self, owner: widgets.Button) -> None:
        for func in self.submission_functions:
            func(owner)

    def _change_selection(self, change=None):
        if self.dropdown.value is not None:
            self.button.description = self.dropdown.value
            self.button.disabled = False
        else:
            self.button.description = "Submit."
            self.button.disabled = True

        self.children = [
            widgets.HBox([self.dropdown, self.button]),
            self.hints[self.dropdown.value],
        ]

    @traitlets.validate("options")
    def _check_options(self, proposal):
        seen = set()
        return [x for x in proposal["value"] if not (x in seen or seen.add(x))]



from collections import OrderedDict, Counter
import pytest

import numpy as np
import pandas as pd

from hypothesis import given, settings
from hypothesis.strategies import (
    booleans,
    floats,
    integers,
    lists,
    tuples,
    one_of,
    sampled_from,
    text,
    composite,
)
from hypothesis.extra.pandas import data_frames, column
from hypothesis.extra.numpy import (
    arrays,
    scalar_dtypes,
    unsigned_integer_dtypes,
    datetime64_dtypes,
    floating_dtypes,
    integer_dtypes,
)

from hypothesis import HealthCheck

from superintendent.queueing import SimpleLabellingQueue


guaranteed_dtypes = one_of(
    scalar_dtypes(),
    unsigned_integer_dtypes(),
    datetime64_dtypes(),
    floating_dtypes(),
    integer_dtypes(),
)


@composite
def dataframe(draw):
    n_cols = draw(integers(min_value=1, max_value=20))
    dtypes = draw(
        lists(
            sampled_from([float, int, str]), min_size=n_cols, max_size=n_cols
        )
    )
    colnames = draw(
        lists(
            text() | integers(), min_size=n_cols, max_size=n_cols, unique=True
        )
    )
    return draw(
        data_frames(
            columns=[
                column(name=name, dtype=dtype)
                for dtype, name in zip(dtypes, colnames)
            ]
        )
    )


@given(input_=one_of(booleans(), floats(), integers(), text()))
def test_enqueueing_and_popping(input_):
    q = SimpleLabellingQueue()
    for i in range(10):
        q.enqueue(input_)
        id_, datapoint = q.pop()
        assert id_ == i
        assert datapoint == input_ or (
            np.isnan(input_) and np.isnan(datapoint)
        )


@given(inputs=lists(one_of(booleans(), floats(), integers(), text())))
def test_enqueue_many(inputs):
    n = len(inputs)
    q = SimpleLabellingQueue()
    q.enqueue_many(inputs)
    assert len(q.data) == n
    # assert we can pop everything:
    for _ in range(n):
        q.pop()
    # assert there's nothing else to pop:
    with pytest.raises(IndexError):
        q.pop()


@settings(suppress_health_check=(HealthCheck.too_slow,))
@given(inputs=dataframe())
def test_enqueue_dataframe(inputs):
    n = len(inputs)
    q = SimpleLabellingQueue()
    q.enqueue_many(inputs)
    assert len(q.data) == n
    # assert we can pop everything:
    for _ in range(n):
        id_, val = q.pop()
        assert isinstance(val, pd.Series)
    # assert there's nothing else to pop:
    with pytest.raises(IndexError):
        q.pop()
    # assert it re-constructs a df on list all
    if n > 0:
        ids, X, y = q.list_all()
        assert isinstance(X, pd.DataFrame) or len(X) == 0
        # assert it re-constructs a df on list uncomplete
        q.submit(ids[0], "hello")
        ids, X = q.list_uncompleted()
        assert isinstance(X, pd.DataFrame) or len(X) == 0
        # assert it re-constructs a df on list uncomplete
        ids, X, y = q.list_completed()
        assert isinstance(X, pd.DataFrame) or len(X) == 0


@settings(suppress_health_check=(HealthCheck.too_slow,))
@given(
    inputs=arrays(
        guaranteed_dtypes,
        tuples(
            integers(min_value=1, max_value=50),
            integers(min_value=1, max_value=50),
        ),
    )
)
def test_enqueue_array(inputs):
    n = inputs.shape[0]
    q = SimpleLabellingQueue()
    q.enqueue_many(inputs)
    assert len(q.data) == n
    # assert we can pop everything:
    for _ in range(n):
        id_, val = q.pop()
        assert isinstance(val, np.ndarray)
        assert len(val.shape) == 1
        assert val.size == inputs.shape[-1]
    # assert there's nothing else to pop:
    with pytest.raises(IndexError):
        q.pop()
    # assert it re-constructs a df on list all
    if n > 0:
        ids, X, y = q.list_all()
        assert isinstance(X, np.ndarray)


@given(
    inputs=lists(one_of(booleans(), floats(), integers(), text())),
    labels=lists(text()),
)
def test_enqueue_with_labels(inputs, labels):

    if len(labels) > len(inputs):
        labels = labels[: len(inputs)]
    elif len(labels) < len(inputs):
        labels += [None] * (len(inputs) - len(labels))

    n = len(inputs) - (len(labels) - Counter(labels)[None])

    q = SimpleLabellingQueue()
    q.enqueue_many(inputs, labels)

    assert len(q.data) == len(inputs)
    # assert we can pop everything where the label was not none:
    for _ in range(n):
        q.pop()
    # assert there's nothing else to pop:
    with pytest.raises(IndexError):
        q.pop()


@given(inputs=lists(one_of(booleans(), floats(), integers(), text())))
def test_enqueue_at_creation(inputs):
    n = len(inputs)
    q = SimpleLabellingQueue(inputs)
    assert len(q.data) == n
    # assert we can pop everything:
    for _ in range(n):
        q.pop()
    # assert there's nothing else to pop:
    with pytest.raises(IndexError):
        q.pop()


@given(label1=text(), label2=text())
def test_submitting_text(label1, label2):
    q = SimpleLabellingQueue()
    q.enqueue(1)
    q.enqueue(2)
    with pytest.raises(ValueError):
        q.submit(0, label1)
    id_, val = q.pop()
    q.submit(id_, label1)
    assert q.progress == 0.5
    id_, val = q.pop()
    q.submit(id_, label2)
    assert q.progress == 1
    assert q.list_labels() == {label1, label2}


@given(label1=text(), label2=text())
def test_submitting_list(label1, label2):
    q = SimpleLabellingQueue()
    q.enqueue(1)
    with pytest.raises(ValueError):
        q.submit(0, label1)
    id_, val = q.pop()
    q.submit(id_, [label1, label2])
    assert q.list_labels() == {label1, label2}


def test_reordering():
    inp = ["b", "a", "d", "c"]
    q = SimpleLabellingQueue()
    q.enqueue_many(inp)
    q.reorder(OrderedDict([(0, 1), (1, 0), (2, 3), (3, 2)]))

    id_, val = q.pop()
    assert val == "a" and id_ == 1
    id_, val = q.pop()
    assert val == "b" and id_ == 0
    id_, val = q.pop()
    assert val == "c" and id_ == 3
    id_, val = q.pop()
    assert val == "d" and id_ == 2


def test_iterating_over_queue():
    inps = [str(i) for i in range(50)]
    q = SimpleLabellingQueue()
    q.enqueue_many(inps)

    for i, (id_, val) in enumerate(q):
        assert i == id_


def test_length_of_queue():
    inps = [str(i) for i in range(50)]
    q = SimpleLabellingQueue()
    assert len(q) == 0
    q.enqueue_many(inps)
    assert len(q) == len(inps)


def test_progress():
    inps = [str(i) for i in range(50)]
    q = SimpleLabellingQueue()
    # assert my little hack for dividing by zero
    assert q.progress == 0
    q.enqueue_many(inps)

    for i, (id_, val) in enumerate(q):
        assert q.progress == i / len(inps)
        q.submit(id_, str(i))


def test_shuffling():
    inps = [str(i) for i in range(50)]
    q = SimpleLabellingQueue()
    q.enqueue_many(inps)
    q.shuffle()
    # assert the order is not the same:
    assert not all([val == inp for inp, (id_, val) in zip(inps, q)])


def test_undo():
    inp = "input 1"
    q = SimpleLabellingQueue()
    q.enqueue(inp)
    id_, val = q.pop()
    q.submit(id_, "label 1")
    # ensure the queue is empty now:
    with pytest.raises(IndexError):
        q.pop()
    q.undo()
    # see if it's possible to pop now:
    id_, val = q.pop()
    assert val == "input 1"


@given(
    inputs=lists(one_of(booleans(), floats(), integers(), text()), min_size=5),
    labels=lists(text(), min_size=5),
)
def test_list_completed(inputs, labels):
    q = SimpleLabellingQueue()
    q.enqueue_many(inputs)

    popped_ids = []
    for i in range(5):
        id_, val = q.pop()
        q.submit(id_, labels[i])

        popped_ids.append(id_)

    ids, x, y = q.list_completed()

    assert len(ids) == 5
    # test that the popped IDs and completed IDs have the same members
    assert pytest.helpers.same_elements(ids, popped_ids)
    assert pytest.helpers.same_elements(y, labels[:5])


@given(
    inputs=lists(one_of(booleans(), floats(), integers(), text()), min_size=5),
    labels=lists(text(), min_size=5),
)
def test_list_uncompleted(inputs, labels):
    q = SimpleLabellingQueue()
    q.enqueue_many(inputs)

    popped_ids = []
    for i in range(5):
        id_, val = q.pop()
        q.submit(id_, labels[i])

        popped_ids.append(id_)

    ids, x = q.list_uncompleted()

    assert len(ids) == (len(inputs) - 5)
    # test that the popped IDs and completed IDs don't share members
    assert pytest.helpers.no_shared_members(ids, popped_ids)
    assert pytest.helpers.same_elements(x, inputs[5:])


@given(
    inputs=lists(one_of(booleans(), floats(), integers(), text()), min_size=5),
    labels=lists(text(), min_size=5),
)
def test_list_all(inputs, labels):
    q = SimpleLabellingQueue()
    q.enqueue_many(inputs)

    popped_ids = []
    for i in range(5):
        id_, val = q.pop()
        q.submit(id_, labels[i])

        popped_ids.append(id_)

    ids, x, y = q.list_all()

    assert len(ids) == len(inputs)
    assert all([label in labels for label in y if label is not None])
    assert all(
        [label is None or id_ in popped_ids for id_, label in zip(ids, y)]
    )
    assert Counter(y)[None] == (len(inputs) - 5)
    assert pytest.helpers.same_elements(ids, range(len(inputs)))



"""Tools to supervise classification."""
import functools
from typing import Callable, Any, Tuple, Optional

from sklearn.preprocessing import MultiLabelBinarizer

from .base import Labeller
from .controls import MulticlassSubmitter
from .queueing import SimpleLabellingQueue


class MultiClassLabeller(Labeller):
    """
    A widget for labelling your data.

    This class is designed to label data for (semi-)supervised learning
    algorithms. It allows you to label data. In the future, it will also allow
    you to re-train an algorithm.
    """

    def __init__(
        self,
        features=None,
        labels=None,
        options=(),
        allow_freetext=True,
        hints=None,
        display_func="default",
        model=None,
        eval_method=None,
        acquisition_function=None,
        shuffle_prop=0.1,
        model_preprocess=None,
        display_preprocess=None,
    ):
        """
        A class for labelling your data.

        This class is designed to label data for (semi-)supervised learning
        algorithms. It allows you to label data, periodically re-train your
        algorithm and assess its performance, and determine which data points
        to label next based on your model's predictions.

        Parameters
        ----------
        features : np.ndarray, pd.DataFrame, sequence
            This should be either a numpy array, a pandas dataframe, or any
            other sequence object (e.g. a list). You can also add data later.
        labels : np.array, pd.Series, sequence
            The labels for your data, if you have some already.
        options : Sequence[ str ]
            The class options, as strings. These will be the labels for the
            buttons. E.g. ["cat", "dog"]
        allow_freetext : bool
            Whether to allow users to add options to this widget using
            free-text input.
        hints : dictionary
            A dictionary of "hints", or examples, of each class. It should map
            from class name (matching the "options" argument above) to data.
            Useful if you want to supply an example of e.g. a person's face
            during labelling.
        display_func : str, func, optional
            A function that accepts a single data point and displays it. Any
            return values are ignored. For convenience, you can supply the
            strings "image" and "default", which are defined in
            superintendent.display.
        model : sklearn.base.BaseEstimator
            An sklearn-interface compliant model (that implements `fit`,
            `predict`, `predict_proba` and `score`).
        eval_method : callable
            A function that accepts three arguments - model, x, and y - and
            returns the score of the model. If None,
            sklearn.model_selection.cross_val_score is used.
        acquisition_function : callable
            A function that re-orders data points during active learning. This
            can be a function that accepts a numpy array (class probabilities)
            or a string referring to a function from
            superintendent.acquisition_functions.
        shuffle_prop : float
            The proportion of data points that is shuffled when re-ordering
            during active learning. This is to avoid biasing too much towards
            the model predictions.
        model_preprocess : callable
            A function that accepts x and y data and returns x and y data. y
            can be None (in which it should return x, None) as this function is
            used on the un-labelled data too.
        display_preprocess : callable
            A function that accepts a single data point and pre-processes it.
            For example, if you have MNIST data as a vector of shape (784,),
            this function could re-shape the data to shape (28, 28)
        """

        input_widget = MulticlassSubmitter(
            hint_function=display_func,
            hints=hints,
            options=options,
            other_option=allow_freetext,
            max_buttons=12,
        )

        queue = SimpleLabellingQueue()

        super().__init__(
            features=features,
            labels=labels,
            queue=queue,
            input_widget=input_widget,
            display_func=display_func,
            model=model,
            eval_method=eval_method,
            acquisition_function=acquisition_function,
            shuffle_prop=shuffle_prop,
            model_preprocess=preprocess_wrapper(model_preprocess),
            display_preprocess=display_preprocess,
        )


def preprocess_wrapper(
    fn: Callable[[Any, Optional[Any]], Tuple[Any, Optional[Any]]]
) -> Callable[[Any, Optional[Any]], Tuple[Any, Optional[Any]]]:
    """
    Wrap a preprocessing function to *also* binarise multi-labels.

    Parameters
    ----------
    fn : callable(featuers, labels=None)
        the preprocessing function to wrap.

    Returns
    -------
    callable(featuers, labels=None)
        The wrapped preprocessor.
    """

    if fn is None:
        return preprocess_multioutput

    @functools.wraps(fn)
    def preprocess(X, y=None):
        X, y = fn(X, y)
        X, y = preprocess_multioutput(X, y)
        return X, y

    return preprocess


def preprocess_multioutput(X, y=None):
    """Turn multilabel lists (sequences of sequences) into binary matrices.

    Parameters
    ----------
    X : any
        The features.
    y : List[List[str]], optional
        The labels - e.g. [['a'], ['a', 'b'], []], with 0 or more labels per
        feature.

    Returns
    -------
    any
        The features, unchanged
    np.ndarray
        A binary matrix of labels.
    """

    if y is None:
        return X, None

    y = MultiLabelBinarizer().fit_transform(y)
    return X, y



"""
Functions to prioritise labelling data points (to drive active learning).
"""
from typing import Dict, Callable
import numpy as np
import scipy.stats

from .decorators import make_acquisition_function

__all__ = ["entropy", "margin", "certainty"]


@make_acquisition_function(handle_multioutput=None)  # noqa: D002
def entropy(probabilities: np.ndarray) -> np.ndarray:
    """
    Sort by the entropy of the probabilities (high to low).

    Parameters
    ----------
    probabilities : np.ndarray
        An array of probabilities, with the shape n_samples,
        n_classes
    shuffle_prop : float
        The proportion of data points that should be randomly shuffled. This
        means the sorting retains some randomness, to avoid biasing your
        new labels and catching any minority classes the algorithm currently
        classifies as a different label.

    """
    neg_entropy = -scipy.stats.entropy(probabilities.T)
    return neg_entropy


@make_acquisition_function(handle_multioutput="mean")  # noqa: D002
def margin(probabilities: np.ndarray) -> np.ndarray:
    """
    Sort by the margin between the top two predictions (low to high).

    Parameters
    ----------
    probabilities : np.ndarray
        An array of probabilities, with the shape n_samples,
        n_classes
    shuffle_prop : float
        The proportion of data points that should be randomly shuffled. This
        means the sorting retains some randomness, to avoid biasing your
        new labels and catching any minority classes the algorithm currently
        classifies as a different label.
    """
    margin = (
        np.sort(probabilities, axis=1)[:, -1]
        - np.sort(probabilities, axis=1)[:, -2]
    )
    return margin


@make_acquisition_function(handle_multioutput="mean")  # noqa: D002
def certainty(probabilities: np.ndarray):
    """
    Sort by the certainty of the maximum prediction.

    Parameters
    ----------
    probabilities : np.ndarray
        An array of probabilities, with the shape n_samples,
        n_classes
    shuffle_prop : float
        The proportion of data points that should be randomly shuffled. This
        means the sorting retains some randomness, to avoid biasing your
        new labels and catching any minority classes the algorithm currently
        classifies as a different label.

    """
    certainty = probabilities.max(axis=-1)
    return certainty


functions: Dict[str, Callable] = {
    "entropy": entropy,
    "margin": margin,
    "certainty": certainty,
}
"""A dictionary of functions to prioritise data."""



import configparser
import itertools
import operator
import warnings
from collections import deque, namedtuple
from contextlib import contextmanager
from datetime import datetime, timedelta
from functools import reduce
from typing import Any, Dict, Sequence, Set, Tuple

import cachetools
import numpy as np
import pandas as pd
import sqlalchemy as sa
import sqlalchemy.ext.declarative
from sqlalchemy.exc import OperationalError, ProgrammingError

from ..queueing.base import BaseLabellingQueue
from ..queueing.utils import _features_to_array
from .serialization import data_dumps, data_loads


def _construct_orm_object(table_name):
    DeclarativeBase = sqlalchemy.ext.declarative.declarative_base()

    class Superintendent(DeclarativeBase):
        __tablename__ = table_name
        id = sa.Column(sa.Integer, primary_key=True)  # noqa: A003
        input = sa.Column(sa.Text)  # noqa: A003
        output = sa.Column(sa.Text, nullable=True)
        inserted_at = sa.Column(sa.DateTime)
        priority = sa.Column(sa.Integer)
        popped_at = sa.Column(sa.DateTime, nullable=True)
        completed_at = sa.Column(sa.DateTime, nullable=True)
        worker_id = sa.Column(sa.String, nullable=True)

    return Superintendent


deserialisers = {"json": data_loads}
serialisers = {"json": data_dumps}


class DatabaseQueue(BaseLabellingQueue):
    """Implements a queue for distributed labelling.

    >>> from superintendent.distributed.dbqueue import Backend
    >>> q = Backend(storage_type='integer_index')
    >>> q.insert(1)
    >>> id_, integer_index = q.pop()
    >>> # ...
    >>> q.submit(id_, value)

    Attributes
    ----------
    data : sqlalchemy.ext.declarative.api.DeclarativeMeta
    deserialiser : builtin_function_or_method
    serialiser : builtin_function_or_method
    """

    worker_id = None
    item = namedtuple("QueueItem", ["id", "data", "label"])

    def __init__(
        self,
        connection_string="sqlite:///:memory:",
        table_name="superintendent",
        storage_type="json",
    ):
        """Instantiate queue for distributed labelling.

        Parameters
        ----------
        connection_string : str, optional
            dialect+driver://username:password@host:port/database. Default:
            'sqlite:///:memory:' (NB: Use only for debugging purposes)
        table_name : str
            The name of the table in SQL where to store the data.
        storage_type : str, optional
            One of 'integer_index', 'pickle' (default) or 'json'.
        """
        self.data = _construct_orm_object(table_name)

        self.deserialiser = deserialisers[storage_type]
        self.serialiser = serialisers[storage_type]
        self.engine = sa.create_engine(connection_string)
        self._popped = deque([])

        if not self.engine.dialect.has_table(
            self.engine, self.data.__tablename__
        ):
            self.data.metadata.create_all(bind=self.engine)

        try:
            # create index for priority
            ix_labelling = sa.Index("ix_labelling", self.data.priority)
            ix_labelling.create(self.engine)
        except OperationalError:
            pass
        except ProgrammingError:
            pass

    @classmethod
    def from_config_file(cls, config_path):
        """Instantiate with database credentials from a configuration file.

        The config file should be an INI file with the following contents:

        [database]
        ; dialect+driver://username:password@host:port/database

        dialect=xxx
        driver=xxx
        username=xxx
        password=xxx
        host=xxx
        port=xxx
        database=xxx

        Parameters
        ----------
        config_path : str
            Path to database configuration file.
        """
        config = configparser.ConfigParser()
        config.read(config_path)

        connection_string_template = (
            "{dialect}+{driver}://"
            "{username}:{password}@{host}:{port}/{database}"
        )
        connection_string = connection_string_template.format(
            **config["database"]
        )
        return cls(connection_string)

    @contextmanager
    def session(self):
        session = sa.orm.Session(bind=self.engine)
        try:
            yield session
            session.commit()
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()

    def enqueue(self, feature, label=None, priority=None):
        """Add a feature to the queue.

        Parameters
        ----------
        feature : Any
            The feature to add.
        label : Any, optional
            The label for the feature.
        priority : int, optional
            The priority of this label in relation to all other priorities in
            the queue.
        """

        now = datetime.now()
        with self.session() as session:
            session.add(
                self.data(
                    input=self.serialiser(feature),
                    inserted_at=now,
                    priority=priority,
                    output=label,
                    completed_at=None if label is None else now,
                )
            )

    def enqueue_many(self, features, labels=None, priorities=None):
        """
        Add items to the queue.

        Parameters
        ----------
        features : Any
            The features to add.
        labels : Any, optional
            The labels for the features.
        priorities : Sequence[int], optional
            The priorities of this label in relation to all other priorities in
            the queue.
        """
        now = datetime.now()
        if isinstance(features, pd.DataFrame):
            features = [row for _, row in features.iterrows()]

        with self.session() as session:
            if priorities is None:
                priorities = itertools.cycle([None])
            if labels is None:
                labels = itertools.cycle([None])

            for feature, label, priority in zip(features, labels, priorities):
                session.add(
                    self.data(
                        input=self.serialiser(feature),
                        inserted_at=now,
                        priority=priority,
                        output=self.serialiser(label),
                        completed_at=None if label is None else now,
                    )
                )

    def reorder(self, priorities: Dict[int, int]) -> None:
        """Re-assign priorities for labels.

        Parameters
        ----------
        priorities : Dict[int, int]
            A mapping from id -> priority.
        """

        self.set_priorities(
            [int(id_) for id_ in priorities.keys()],
            [int(priority) for priority in priorities.values()],
        )

    def set_priorities(self, ids: Sequence[int], priorities: Sequence[int]):
        """Set the priorities for data points.

        Parameters
        ----------
        ids : Sequence[int]
            The IDs for which to change the priority.
        priorities : Sequence[int]
            The priorities.
        """

        with self.session() as session:
            rows = session.query(self.data).filter(self.data.id.in_(ids)).all()
            for row in rows:
                row.priority = priorities[ids.index(row.id)]

    def pop(self, timeout: int = 600) -> Tuple[int, Any]:
        """Pop an item from the queue.

        Parameters
        ----------
        timeout : int
            How long ago an item must have been popped in order for it to be
            popped again.

        Raises
        ------
        IndexError
            If there are no more items to pop.

        Returns
        -------
        id : int
            The ID of the popped data point
        data : Any
            The datapoint.
        """

        with self.session() as session:
            row = (
                session.query(self.data)
                .filter(
                    self.data.completed_at.is_(None)
                    & (
                        self.data.popped_at.is_(None)
                        | (
                            self.data.popped_at
                            < (datetime.now() - timedelta(seconds=timeout))
                        )
                    )
                )
                .order_by(self.data.priority)
                .first()
            )
            if row is None:
                raise IndexError("Trying to pop off an empty queue.")
            else:
                row.popped_at = datetime.now()
                id_ = row.id
                value = row.input
                self._popped.append(id_)
                return id_, self.deserialiser(value)

    def submit(self, id_: int, label: str) -> None:
        """Submit a label for a data point.

        Parameters
        ----------
        id_ : int
            The ID for which you are submitting a data point.
        label : str
            The label you want to submit.

        Raises
        ------
        ValueError
            If you haven't popped an item yet.
        """

        if id_ not in self._popped:
            raise ValueError("This item was not popped; you cannot label it.")
        with self.session() as session:
            row = session.query(self.data).filter_by(id=id_).first()
            row.output = self.serialiser(label)
            row.worker_id = self.worker_id
            row.completed_at = datetime.now()

    def undo(self) -> None:
        """Undo the most recently popped item."""

        if len(self._popped) > 0:
            id_ = self._popped.pop()
            self._reset(id_)

    def _reset(self, id_: int) -> None:
        with self.session() as session:
            row = session.query(self.data).filter_by(id=id_).first()
            row.output = None
            row.completed_at = None
            row.popped_at = None

    def list_all(self):
        with self.session() as session:
            objects = session.query(self.data).all()

            items = [
                self.item(
                    id=obj.id,
                    data=self.deserialiser(obj.input),
                    label=self.deserialiser(obj.output),
                )
                for obj in objects
            ]
        ids = [item.id for item in items]
        x = _features_to_array([item.data for item in items])
        y = [item.label for item in items]

        return ids, x, y

    def list_completed(self):
        with self.session() as session:
            objects = (
                session.query(self.data)
                .filter(
                    self.data.output.isnot(None)
                    & self.data.completed_at.isnot(None)
                )
                .all()
            )

            items = [
                self.item(
                    id=obj.id,
                    data=self.deserialiser(obj.input),
                    label=self.deserialiser(obj.output),
                )
                for obj in objects
            ]

        ids = [item.id for item in items]
        x = _features_to_array([item.data for item in items])
        y = [item.label for item in items]

        return ids, x, y

    def list_labels(self) -> Set[str]:
        with self.session() as session:
            rows = (
                session.query(self.data.output)
                .filter(self.data.output.isnot(None))
                .distinct()
            )
            try:
                return set([self.deserialiser(row.output) for row in rows])
            except TypeError:
                return reduce(
                    operator.or_,
                    [
                        set(self.deserialiser(row.output))
                        if row.output is not None
                        else set()
                        for row in rows
                    ],
                )

    def list_uncompleted(self):
        with self.session() as session:
            objects = (
                session.query(self.data)
                .filter(self.data.output.is_(None))
                .all()
            )
            items = [
                self.item(
                    id=obj.id,
                    data=self.deserialiser(obj.input),
                    label=obj.output,
                )
                for obj in objects
            ]

            ids = [obj.id for obj in objects]
            x = _features_to_array([item.data for item in items])

            return ids, x

    def clear_queue(self):
        self._popped = deque([])
        with self.session() as session:
            session.query(self.data).delete()

    def drop_table(self, sure=False):  # noqa: D001
        if sure:
            self.data.metadata.drop_all(bind=self.engine)
        else:
            warnings.warn("To actually drop the table, pass sure=True")

    def _unlabelled_count(self) -> int:
        with self.session() as session:
            return (
                session.query(self.data)
                .filter(
                    self.data.completed_at.is_(None)
                    & self.data.output.is_(None)
                )
                .count()
            )

    def _labelled_count(self) -> int:
        with self.session() as session:
            return (
                session.query(self.data)
                .filter(
                    self.data.completed_at.isnot(None)
                    & self.data.output.isnot(None)
                )
                .count()
            )

    @cachetools.cached(cachetools.TTLCache(1, 60))
    def _total_count(self):
        with self.session() as session:
            n_total = session.query(self.data).count()
        return n_total

    @property
    def progress(self) -> float:
        try:
            return self._labelled_count() / self._total_count()
        except ZeroDivisionError:
            return np.nan

    def __len__(self):
        with self.session() as session:
            return (
                session.query(self.data)
                .filter(self.data.completed_at.is_(None))
                .count()
            )

    def __iter__(self):
        return self

    def __next__(self):
        try:
            return self.pop()
        except IndexError:
            raise StopIteration



import time
from math import inf
from typing import Optional

import ipywidgets as widgets
import traitlets

from .._compatibility import ignore_widget_on_submit_warning
from .queueing import DatabaseQueue


class _DistributedMixin:

    worker_id = traitlets.Unicode(allow_none=True)

    def __init__(
        self,
        *,
        connection_string="sqlite:///:memory:",
        worker_id=False,
        table_name="superintendent",
        **kwargs
    ):

        super().__init__(**kwargs)

        # override the queue:
        self.queue = DatabaseQueue(
            connection_string=connection_string, table_name=table_name
        )

        if kwargs.get("features") is not None:
            self.add_features(kwargs.get("features"), kwargs.get("labels"))

        if worker_id and not isinstance(worker_id, str):
            self._get_worker_id()
            return
        else:
            self.queue.worker_id = worker_id if worker_id else None
            self._annotation_loop = self._annotation_iterator()
            self.queue.undo()
            next(self._annotation_loop)

    def _get_worker_id(self):
        worker_id_field = widgets.Text(
            placeholder="Please enter your name or ID."
        )
        self.layout.children = [
            widgets.HTML(value="<h2>Please enter your name:</h2>"),
            widgets.Box(
                children=[worker_id_field],
                layout=widgets.Layout(
                    justify_content="center",
                    padding="5% 0",
                    display="flex",
                    width="100%",
                    min_height="150px",
                ),
            ),
        ]
        with ignore_widget_on_submit_warning():
            worker_id_field.on_submit(self._set_worker_id)

    def _set_worker_id(self, worker_id_field):
        if len(worker_id_field.value) > 0:
            self.queue.worker_id = worker_id_field.value
            self._compose()
        self._annotation_loop = self._annotation_iterator()
        self.queue.undo()
        next(self._annotation_loop)

    def orchestrate(
        self,
        interval_seconds: Optional[float] = None,
        interval_n_labels: int = 0,
        shuffle_prop: float = 0.1,
        max_runs: float = inf,
    ):
        """Orchestrate the active learning process.

        This method can either re-train the classifier and re-order the data
        once, or it can run a never-ending loop to re-train the model at
        regular intervals, both in time and in the size of labelled data.

        Parameters
        ----------
        interval_seconds : int, optional
            How often the retraining should occur, in seconds. If this is None,
            the retraining only happens once, then returns (this is suitable)
            if you want the retraining schedule to be maintained e.g. by a cron
            job). The default is 60 seconds.
        interval_n_labels : int, optional
            How many new data points need to have been labelled in between runs
            in order for the re-training to occur.
        shuffle_prop : float
            What proportion of the data should be randomly sampled on each re-
            training run.
        max_runs : float, int
            How many orchestration runs to do at most. By default infinite.

        Returns
        -------
        None
        """
        if interval_seconds is None:
            self._run_orchestration(
                interval_seconds=0,
                interval_n_labels=interval_n_labels,
                shuffle_prop=shuffle_prop,
            )
        else:
            runs = 0
            while runs < max_runs:
                runs += self._run_orchestration(
                    interval_seconds=interval_seconds,
                    interval_n_labels=interval_n_labels,
                    shuffle_prop=shuffle_prop,
                )

    def _run_orchestration(
        self,
        interval_seconds: float = 60,
        interval_n_labels: int = 0,
        shuffle_prop: float = 0.1,
    ) -> bool:

        first_orchestration = not hasattr(self, "_last_n_labelled")

        if first_orchestration:
            self._last_n_labelled = 0

        n_new_labels = self.queue._labelled_count() - self._last_n_labelled
        if first_orchestration or n_new_labels >= interval_n_labels:
            self._last_n_labelled += n_new_labels
            self.shuffle_prop = shuffle_prop
            self.retrain()  # type: ignore
            print(self.model_performance.value)  # type: ignore
            time.sleep(interval_seconds)
            return True
        else:
            return False



import operator
from collections import defaultdict, deque
from functools import reduce
from random import shuffle
from typing import Any, DefaultDict, Deque, Dict, List

import numpy as np
import pandas as pd

from .base import BaseLabellingQueue
from .utils import _features_to_array


class ClusterLabellingQueue(BaseLabellingQueue):
    def __init__(
        self,
        features: Any = None,
        cluster_indices: Any = None,
        representativeness=None,
    ):
        """Create a queue for labelling clusters.

        Parameters
        ----------
        features : Any
            The features you'd like to add to the queue.
        cluster_indices : Any
            The clusters that each of the data points belong to. This should
            match the features in length.
        representativeness : Any, optional
            The respective cluster representativeness of each data point. This
            could be distance from cluster center, probability of cluster
            membership, or a similar metric.
        """

        self.data: DefaultDict[Any, List[Any]] = defaultdict(list)
        self.representativeness: DefaultDict[Any, List[Any]] = defaultdict(
            list
        )
        self.cluster_labels: Dict[Any, str] = dict()

        self.order: Deque[int] = deque([])
        self._popped: Deque[int] = deque([])

        if features is not None:
            self.enqueue_many(features, cluster_indices, representativeness)

    def enqueue_many(self, features, cluster_indices, representativeness=None):
        """Add items to the queue.

        Parameters
        ----------
        features : Any
            The features you'd like to add to the queue.
        cluster_indices : Any
            The clusters that each of the data points belong to. This should
            match the features in length.
        representativeness : Any, optional
            The respective cluster representativeness of each data point. This
            could be distance from cluster center, probability of cluster
            membership, or a similar metric.

        Returns
        -------
        None
        """

        if isinstance(features, pd.DataFrame):
            features = [row for _, row in features.iterrows()]

        if representativeness is None:
            representativeness = np.full(len(features), np.nan)

        for cluster_index, feature, represents in zip(
            cluster_indices, features, representativeness
        ):
            self.enqueue(cluster_index, feature, represents)

    def enqueue(self, cluster_index, feature, representativeness=None):
        """Add an item to the queue.

        Parameters
        ----------
        cluster_index : Any
            The cluster index
        feature : Any
            The data to be added to the queue.
        representativeness : float, optional
            The respective representativeness of the data point. This
            could be distance from cluster center, probability of cluster
            membership, or a similar metric. (the default is None)
        """
        self.data[cluster_index].append(feature)
        if representativeness is None:
            representativeness = np.nan
        self.representativeness[cluster_index].append(representativeness)

        if cluster_index not in self.order:
            self.order.appendleft(cluster_index)

    def pop(self):
        """Pop an item off the queue.

        Returns
        -------
        id_ : int
            The ID of the cluster.
        features : Any
            The data points that are in this cluster.
        """

        id_ = self.order.pop()
        self._popped.append(id_)
        features = [
            x
            for _, x in sorted(
                zip(self.representativeness[id_], self.data[id_]),
                key=lambda pair: pair[0],
            )
        ]
        return id_, _features_to_array(features)

    def submit(self, cluster_index, cluster_label):
        """Submit a label for a cluster.

        Parameters
        ----------
        cluster_index : Any
            The cluster you are trying to label.
        cluster_label : str
            The label for the cluster

        Raises
        ------
        ValueError
            If you are trying to label a cluster you haven't popped off the
            queue.
        """

        if cluster_index not in self._popped:
            raise ValueError("This item was not popped; you cannot label it.")
        self.cluster_labels[cluster_index] = cluster_label

    def reorder(self):
        """Re-order the queue. This is currently not implemented."""
        pass

    def shuffle(self) -> None:
        """Shuffle the queue."""
        _order = list(self.order)
        shuffle(_order)
        self.order = deque(_order)

    def undo(self) -> None:
        """Unpop the most recently popped item."""
        if len(self._popped) > 0:
            cluster_index = self._popped.pop()
            self.cluster_labels.pop(cluster_index, None)
            self.order.append(cluster_index)

    def list_completed(self):
        """List the data that has been assigned a cluster label.

        Returns
        -------
        cluster_indices
            The indices of the clusters.
        features
            The features that have been assigned a label.
        cluster_labels
            The assigned cluster labels.
        """

        features = [
            data
            for idx, values in self.data.items()
            for data in values
            if idx in self.cluster_labels
        ]
        cluster_indices = [
            idx
            for idx, values in self.data.items()
            for data in values
            if idx in self.cluster_labels
        ]
        cluster_labels = [
            self.cluster_labels[idx]
            for idx, values in self.data.items()
            for data in values
            if idx in self.cluster_labels
        ]

        return cluster_indices, _features_to_array(features), cluster_labels

    def list_uncompleted(self):
        """List the data that has not yet been assigned a label.

        Returns
        -------
        cluster_indices
            The indices of the clusters the data points are in.
        features
            The data in the unlabelled features.
        """

        features = [
            data
            for idx, values in self.data.items()
            for data in values
            if idx not in self.cluster_labels
        ]
        cluster_indices = [
            idx
            for idx, values in self.data.items()
            for data in values
            if idx not in self.cluster_labels
        ]

        return cluster_indices, _features_to_array(features)

    def list_all(self):
        """List all data.

        Returns
        -------
        cluster_indices
            The indices of the clusters the data points are in.
        features
            The data.
        cluster_labels
            The assigned cluster labels.
        """

        features = [
            data for idx, values in self.data.items() for data in values
        ]
        cluster_indices = [
            idx for idx, values in self.data.items() for data in values
        ]
        cluster_labels = [
            self.cluster_labels.get(idx)
            for idx, values in self.data.items()
            for data in values
        ]

        return cluster_indices, _features_to_array(features), cluster_labels

    @property
    def progress(self):
        """How much of the queue has been completed.

        Returns
        -------
        progress : float
            The progress.
        """

        try:
            return len(self.cluster_labels) / len(self.data)
        except ZeroDivisionError:
            return np.nan

    def list_labels(self):
        try:
            return set(sorted(self.cluster_labels.values()))
        except TypeError:
            return reduce(operator.or_, map(set, self.cluster_labels.values()))

    def __iter__(self):
        return self

    def __len__(self):
        return len(self.order)

    def __next__(self):
        try:
            return self.pop()
        except IndexError:
            raise StopIteration



import itertools
import operator
from collections import deque, namedtuple
from functools import reduce
from random import shuffle
from typing import Any, Deque, Dict, Optional, Set, Tuple

import pandas as pd

from .base import BaseLabellingQueue
from .utils import _features_to_array


class SimpleLabellingQueue(BaseLabellingQueue):

    item = namedtuple("QueueItem", ["id", "data", "label"])

    def __init__(self, features: Any = None, labels: Any = None):
        """Create an in-memory labelling queue.

        Parameters
        ----------
        features : Any, optional
            Features to be added to the queue. You can either provide them
            here, or later using the enqueue_many method (the default is None).
        labels : Any, optional
            Labels for the features to be added to the queue. You can either
            provide them here, or later using the enqueue_many method
            (the default is None).
        """

        self.data: Dict[int, Any] = dict()
        self.labels: Dict[int, Any] = dict()

        self.order: Deque[int] = deque([])
        self._popped: Deque[int] = deque([])

        if features is not None:
            self.enqueue_many(features, labels)

    def enqueue(self, feature: Any, label: Optional[Any] = None) -> None:
        """Add a data point to the queue.

        Parameters
        ----------
        feature : Any
            A data point to be added to the queue
        label : str, list, optional
            The label, if you already have one (the default is None)

        Returns
        -------
        None
        """

        if len(self.data) > 0:
            idx = max(self.data.keys()) + 1
        else:
            idx = 0

        self.data[idx] = feature

        if label is not None:
            self.labels[idx] = label
        else:
            self.order.appendleft(idx)

    def enqueue_many(
        self, features: Any, labels: Optional[Any] = None
    ) -> None:
        """Add a bunch of items to the queue.

        Parameters
        ----------
        features : Any
            [description]
        labels : [type], optional
            [description] (the default is None, which [default_description])

        Returns
        -------
        None
            [description]
        """

        if isinstance(features, pd.DataFrame):
            features = [row for _, row in features.iterrows()]

        if labels is None:
            labels = itertools.cycle([None])

        for feature, label in zip(features, labels):
            self.enqueue(feature, label)

    def pop(self) -> Tuple[int, Any]:
        """Pop an item off the queue.

        Returns
        -------
        int
            The ID of the item you just popped
        Any
            The item itself.
        """

        id_ = self.order.pop()
        self._popped.append(id_)
        return id_, self.data[id_]

    def submit(self, id_: int, label: Any) -> None:
        """Label a data point.

        Parameters
        ----------
        id_ : int
            The ID of the datapoint to submit a label for
        label : str
            The label to apply for the data point

        Raises
        ------
        ValueError
            If you attempt to label an item that hasn't been popped in this
            queue.

        Returns
        -------
        None
        """

        if id_ not in self._popped:
            raise ValueError("This item was not popped; you cannot label it.")
        self.labels[id_] = label

    def reorder(self, new_order: Dict[int, int]) -> None:
        """Reorder the data still in the queue

        Parameters
        ----------
        new_order : Dict[int, int]
            A mapping from ID of an item to the order of the item. For example,
            a dictionary {1: 2, 2: 1, 3: 3} would place the item with ID 2
            first, then the item with id 1, then the item with ID 3.

        Returns
        -------
        None
        """
        self.order = deque(
            [
                idx
                for idx, _ in sorted(
                    new_order.items(), key=lambda item: -item[1]
                )
            ]
        )

    def shuffle(self) -> None:
        """Shuffle the queue.

        Returns
        -------
        None
        """
        _order = list(self.order)
        shuffle(_order)
        self.order = deque(_order)

    def undo(self) -> None:
        """Un-pop the latest item.

        Returns
        -------
        None
        """

        if len(self._popped) > 0:
            id_ = self._popped.pop()
            self.labels.pop(id_, None)
            self.order.append(id_)

    def list_completed(self):
        """List all items with a label.

        Returns
        -------
        ids : List[int]
            The IDs of the returned items.
        x : Any
            The data points that have labels.
        y : Any
            The labels.
        """

        items = [
            self.item(id=id_, data=self.data[id_], label=self.labels.get(id_))
            for id_ in sorted(self._popped)
            if id_ in self.labels
        ]
        ids = [item.id for item in items]
        x = _features_to_array([item.data for item in items])
        y = [item.label for item in items]
        return ids, x, y

    def list_uncompleted(self):
        """List all items without a label.

        Returns
        -------
        ids : List[int]
            The IDs of the returned items.
        x : Any
            The data points that don't have labels.
        """

        items = [
            self.item(id=id_, data=self.data[id_], label=None)
            for id_ in sorted(self.order)
            if id_ not in self.labels
        ]
        ids = [item.id for item in items]
        x = _features_to_array([item.data for item in items])
        return ids, x

    def list_all(self):
        """List all items.

        Returns
        -------
        ids : List[int]
            The IDs of the returned items.
        x : Any
            The data points.
        y : Any
            The labels.
        """

        items = [
            self.item(id=id_, data=self.data[id_], label=self.labels.get(id_))
            for id_ in self.data
        ]
        ids = [item.id for item in items]
        x = _features_to_array([item.data for item in items])
        y = [item.label for item in items]
        return ids, x, y

    def list_labels(self) -> Set[str]:
        """List all the labels.

        Returns
        -------
        Set[str]
            All the labels.
        """

        try:
            return set(sorted(self.labels.values()))
        except TypeError:
            return reduce(operator.or_, map(set, self.labels.values()))

    @property
    def progress(self) -> float:
        """The queue progress."""

        if len(self.data) > 0:
            return len(self.labels) / len(self.data)
        else:
            return 0

    def __len__(self):
        return len(self.order)

    def __iter__(self):
        return self

    def __next__(self):
        try:
            return self.pop()
        except IndexError:
            raise StopIteration



import abc
from typing import Any, Dict, Optional, Tuple


class BaseLabellingQueue(abc.ABC):  # pragma: no cover
    @abc.abstractmethod
    def enqueue(self, feature: Any, label: Optional[Any] = None):
        """Add a data point to the queue.

        Parameters
        ----------
        feature : Any
            A data point to be added to the queue
        label : str, list, optional
            The label, if you already have one (the default is None)

        Returns
        -------
        None
        """
        pass

    @abc.abstractmethod
    def enqueue_many(self, features: Any, labels: Optional[Any] = None):
        """Add multiple data points to the queue.

        Parameters
        ----------
        features : Any
            A set of data points to be added to the queue.
        labels : str, list, optional
            The labels for this data point.

        Returns
        -------
        None
        """
        pass

    @abc.abstractmethod
    def pop(self) -> Tuple[int, Any]:
        """Pop an item off the queue.

        Returns
        -------
        int
            The ID of the item just popped
        Any
            The item itself.
        """
        pass

    @abc.abstractmethod
    def submit(self, id_: int, label: str) -> None:
        """Label a data point.

        Parameters
        ----------
        id_ : int
            The ID of the datapoint to submit a label for
        label : str
            The label to apply for the data point

        Raises
        ------
        ValueError
            If you attempt to label an item that hasn't been popped in this
            queue.

        Returns
        -------
        None
        """
        pass

    @abc.abstractmethod
    def reorder(self, new_order: Dict[int, int]) -> None:
        """Reorder the data still in the queue

        Parameters
        ----------
        new_order : Dict[int, int]
            A mapping from ID of an item to the order of the item. For example,
            a dictionary {1: 2, 2: 1, 3: 3} would place the item with ID 2
            first, then the item with id 1, then the item with ID 3.

        Returns
        -------
        None
        """
        pass

    @abc.abstractmethod
    def undo(self) -> None:
        """Un-pop the latest item.

        Returns
        -------
        None
        """
        pass

    @abc.abstractmethod
    def list_completed(self):
        """List all items with a label.

        Returns
        -------
        ids : List[int]
            The IDs of the returned items.
        x : Any
            The data points that have labels.
        y : Any
            The labels.
        """
        pass

    @abc.abstractmethod
    def list_uncompleted(self):
        """List all items without a label.

        Returns
        -------
        ids : List[int]
            The IDs of the returned items.
        x : Any
            The data points that don't have labels.
        """
        pass

    @abc.abstractmethod
    def list_labels(self):
        """List all the labels.

        Returns
        -------
        Set[str]
            All the labels.
        """
        pass

    @abc.abstractmethod
    def __iter__(self):
        pass

    @abc.abstractmethod
    def __next__(self):
        pass



import pytest

import ipywidgets

from hypothesis import strategies as st
from hypothesis import given, settings

from superintendent import controls
from superintendent.controls import buttongroup
from superintendent.controls import dropdownbutton


DUMMY_KEYUP_DICT_ONE = {
    "altKey": False,
    "ctrlKey": False,
    "metaKey": False,
    "shiftKey": False,
    "type": "keyup",
    "timeStamp": 129242.40000001737,
    "code": "Digit1",
    "key": "1",
    "location": 0,
    "repeat": False,
    "event": "keyup",
}


@settings(deadline=None)
@given(options=st.lists(st.text()))
def test_that_options_are_set_correctly_when_instantiated(options):
    widget = controls.Submitter(options=options)
    assert widget.options == options


@settings(deadline=None)
@given(options=st.lists(st.text()))
def test_that_options_are_set_correctly_when_updated(options):
    widget = controls.Submitter()
    assert widget.options == []

    widget.options = options
    assert widget.options == options


@settings(deadline=None)
@given(options=st.lists(st.integers()))
def test_that_options_are_converted_to_strings(options):
    widget = controls.Submitter(options=options)
    assert widget.options == [str(option) for option in options]


def test_that_updating_options_triggers_compose(mocker):
    options = ["a", "b", "c"]
    widget = controls.Submitter()
    mock_compose = mocker.patch.object(widget, "_compose")
    widget.options = options
    mock_compose.assert_called_once()


def test_that_on_submit_updates_submissions_functions():
    widget = controls.Submitter()
    widget.on_submit(print)
    assert print in widget.submission_functions


def test_that_on_submit_fails_when_not_given_callable():
    widget = controls.Submitter()
    with pytest.raises(ValueError):
        widget.on_submit("dummy")


def test_that_submit_passes_correct_values(mocker):
    widget = controls.Submitter(["a", "b"])
    mock_function = mocker.Mock()
    widget.on_submit(mock_function)

    widget.other_widget.value = "dummy submission"
    widget._submit(widget.other_widget)
    assert mock_function.call_args[0] == ("dummy submission",)

    widget._submit(list(widget.control_elements.buttons.values())[0])
    assert mock_function.call_args[0] == ("a",)


def test_on_undo_args_get_called(mocker):
    widget = controls.Submitter(["a", "b"])
    mock_undo = mocker.MagicMock()
    widget.on_undo(mock_undo)
    widget._undo("dummy")

    mock_undo.assert_called_once()
    assert mock_undo.call_args[0] == tuple()


def test_on_skip_args_get_called(mocker):
    widget = controls.Submitter(["a", "b"])
    mock_skip = mocker.MagicMock()
    widget.on_skip(mock_skip)
    widget._skip("dummy")

    mock_skip.assert_called_once()
    assert mock_skip.call_args[0] == tuple()


def test_that_sort_options_sorts_options(mocker):
    options = ["d", "a", "b", "c"]
    widget = controls.Submitter(options)
    mock_compose = mocker.patch.object(widget, "_compose")
    widget._sort_options()

    assert widget.options == ["a", "b", "c", "d"]
    mock_compose.assert_called_once()


def test_that_removing_core_options_fails():
    options = ["a", "b", "c"]
    widget = controls.Submitter(options)

    widget.remove_options(["a"])

    assert widget.options == options


def test_that_removing_added_options_works():
    options = ["a", "b", "c"]
    widget = controls.Submitter(options)

    widget.options += ["d"]
    assert widget.options == options + ["d"]

    widget.remove_options(["d"])
    assert widget.options == options


def test_that_submitting_a_new_option_adds_it(mocker):
    widget = controls.Submitter(["a", "b"])
    mock_function = mocker.Mock()
    widget.on_submit(mock_function)

    widget.other_widget.value = "dummy submission"
    widget._submit(widget.other_widget)

    assert "dummy submission" in widget.options


def test_that_a_button_group_is_created():
    options = ["a", "b", "c"]
    widget = controls.Submitter(options)
    assert isinstance(widget.control_elements, buttongroup.ButtonGroup)
    assert widget.options == widget.control_elements.options
    widget.options = ["a", "b", "c", "d"]
    assert widget.options == widget.control_elements.options


def test_that_a_button_dropdown_is_created():
    options = ["a", "b", "c", "d", "e", "f"]
    widget = controls.Submitter(options, max_buttons=3)
    assert isinstance(widget.control_elements, dropdownbutton.DropdownButton)
    assert widget.options == widget.control_elements.options
    widget.options = ["a", "b", "c", "d"]
    assert widget.options == widget.control_elements.options


def test_that_by_default_other_option_is_enabled():
    options = ["a", "b", "c", "d", "e", "f"]
    widget = controls.Submitter(options)

    assert isinstance(widget.other_widget, ipywidgets.Text)


def test_that_disabling_other_option_works():
    options = ["a", "b", "c", "d", "e", "f"]
    widget = controls.Submitter(options, other_option=False)

    assert not any(
        isinstance(item, ipywidgets.Text)
        for item in widget.children[-1].children
    )


def test_that_add_hint_calls_hint_function(mocker):
    options = ["a", "b", "c", "d", "e", "f"]

    mock_hint_function = mocker.Mock()

    widget = controls.Submitter(options, hint_function=mock_hint_function)

    assert widget.hint_function is mock_hint_function
    assert mock_hint_function.call_count == 0

    widget.add_hint("a", "test hint")
    assert mock_hint_function.call_count == 1
    assert mock_hint_function.call_args == (("test hint",),)


def test_that_passing_hints_calls_hint_function(mocker):
    options = ["a", "b", "c", "d", "e", "f"]

    mock_hint_function = mocker.Mock()

    widget = controls.Submitter(  # noqa
        options, hint_function=mock_hint_function, hints={"a": "test hint"}
    )

    assert mock_hint_function.call_count == 1
    assert mock_hint_function.call_args == (("test hint",),)



from .. import multiclass_labeller
from .mixin import _DistributedMixin


class MultiClassLabeller(
    _DistributedMixin, multiclass_labeller.MultiClassLabeller
):
    """
    A class for labelling your data.

    This class is designed to label data for (semi-)supervised learning
    algorithms. It allows you to label data. In the future, it will also allow
    you to re-train an algorithm.

    Parameters
    ----------
    connection_string: str
        A SQLAlchemy-compatible database connection string. This is where the
        data for this widget will be stored, and where it will be retrieved
        from for labelling.
    features : list, np.ndarray, pd.Series, pd.DataFrame, optional
        An array or sequence of data in which each element (if 1D) or each row
        (if 2D) represents one data point for which you'd like to generate
        labels.
    labels : list, np.ndarray, pd.Series, pd.DataFrame, optional
        If you already have some labels, but would like to re-label some, then
        you can pass these in as labels.
    options : tuple, list
        The options presented for labelling.
    classifier : sklearn.base.ClassifierMixin, optional
        An object that implements the standard sklearn fit/predict methods. If
        provided, a button for retraining the model is shown, and the model
        performance under k-fold crossvalidation can be read as you go along.
    display_func : callable, optional
        A function that will be used to display the data. This function should
        take in two arguments, first the data to display, and second the number
        of data points to display (set to 1 for this class).
    eval_method : callable, optional
        A function that accepts the classifier, features, and labels as input
        and returns a dictionary of values that contain the key 'test_score'.
        The default is sklearn.model_selection.cross_validate, with cv=3. Use
        functools.partial to create a function with its parameters fixed.
    reorder : str, callable, optional
        One of the reordering algorithms specified in
        :py:mod:`superintendent.prioritisation`. This describes a function that
        receives input in the shape of n_samples, n_labels and calculates the
        priority in terms of information value in labelling a data point.
    shuffle_prop : float
        The proportion of points that are shuffled when the data points are
        re-ordered (see reorder keyword-argument). This controls the
        "exploration vs exploitation" trade-off - the higher, the more you
        explore the feature space randomly, the lower, the more you exploit
        your current weak points.
    keyboard_shortcuts : bool, optional
        If you want to enable ipyevent-mediated keyboard capture to use the
        keyboard rather than the mouse to submit data.

    """

    pass



import numpy as np

from superintendent.acquisition_functions import margin, entropy, certainty

from superintendent.acquisition_functions.decorators import _shuffle_subset


def test_no_shuffling():
    test_data = np.random.rand(100, 10)
    for _ in range(10):
        assert (test_data == _shuffle_subset(test_data, 0)).all()


def test_shuffling_all():
    test_data = np.random.rand(100, 10)
    assert not (test_data == _shuffle_subset(test_data.copy(), 1.0)).all()


def test_margin():
    probabilites = np.array(
        [[0.2, 0.3, 0.1], [0.01, 0.9, 0.09], [0.5, 0.5, 0.0]]
    )
    assert (margin(probabilites, shuffle_prop=0) == np.array([1, 2, 0])).all()


def test_entropy():
    probabilites = np.array(
        [[0.2, 0.3, 0.1], [0.01, 0.9, 0.09], [0.5, 0.5, 0.0]]
    )
    assert (entropy(probabilites, shuffle_prop=0) == np.array([0, 2, 1])).all()


def test_certainty():
    probabilites = np.array(
        [[0.2, 0.3, 0.1], [0.01, 0.9, 0.09], [0.5, 0.5, 0.0]]
    )
    assert (
        certainty(probabilites, shuffle_prop=0) == np.array([0, 2, 1])
    ).all()



import functools
import typing

import numpy as np


def _shuffle_subset(data: np.ndarray, shuffle_prop: float) -> np.ndarray:
    to_shuffle = np.nonzero(np.random.rand(data.shape[0]) < shuffle_prop)[0]
    data[to_shuffle, ...] = data[np.random.permutation(to_shuffle), ...]
    return data


def _get_indices(scores: np.ndarray, shuffle_prop: float) -> np.ndarray:
    """
    Get the indices of each score (lowest first) that sort the array,
    in order for each data point.

    The double argsort might seem confusing, but it makes sense - it's the
    function that gets the location of each index in a sorted version of the
    index.
    """
    return _shuffle_subset(scores.argsort().argsort(), shuffle_prop)


def _is_multioutput(
    probabilities: typing.Union[np.ndarray, typing.List[np.ndarray]]
):
    if isinstance(probabilities, list) and (
        isinstance(probabilities[0], np.ndarray)
        and probabilities[0].ndim == 2
        and np.isclose(probabilities[0].sum(-1), 1).all()
    ):
        return True
    elif isinstance(probabilities, np.ndarray):
        return False
    else:
        raise ValueError("Unknown probability format.")


def make_acquisition_function(handle_multioutput="mean"):
    """Wrap an acquisition function.

    This ensures the function can accept a shuffle proportion and ensures it
    handles multi_output.

    Parameters
    ----------
    handle_multioutput : str, optional
        How the acquisition function should handle multioutput data, which
        comes as a list of binary classifier outputs.
    """

    def decorator(fn):
        if handle_multioutput == "mean":  # define fn where scores are avgd

            @functools.wraps(fn)
            def wrapped_fn(probabilities, shuffle_prop=0.1):
                if _is_multioutput(probabilities):
                    scores = np.stack(
                        tuple(fn(prob) for prob in probabilities), axis=0
                    ).mean(axis=0)
                else:
                    scores = fn(probabilities)
                return _get_indices(scores, shuffle_prop)

        else:  # raise error if list is passed

            @functools.wraps(fn)
            def wrapped_fn(probabilities, shuffle_prop=0.1):
                if _is_multioutput(probabilities):
                    raise ValueError(
                        "The input probabilities is a list of arrays, "
                        + "indicating multi-label output. "
                        + "The {} function ".format(fn.__name__)
                        + "is not defined for these outputs. Use "
                        + "the acquisition functions margin or certainty "
                        + "instead."
                    )
                else:
                    scores = fn(probabilities)
                return _get_indices(scores, shuffle_prop)

        return wrapped_fn

    return decorator



import pytest

pytestmark = pytest.mark.skip

# from sklearn.linear_model import LogisticRegression
# from superintendent.validation import valid_classifier  # , valid_data


# def test_valid_classifier():
#     assert valid_classifier(None) is None
#     dummy_classifier = LogisticRegression()
#     assert valid_classifier(dummy_classifier) is dummy_classifier
#     with pytest.raises(ValueError):
#         valid_classifier("dummy-non-classifier")


# def test_valid_data():
#     assert valid_classifier(None) is None
#     dummy_classifier = LogisticRegression()
#     assert valid_classifier(dummy_classifier) is dummy_classifier
#     with pytest.raises(ValueError):
#         valid_classifier("dummy-non-classifier")



import time

import numpy as np

import pytest

from superintendent.base import Labeller
from superintendent import display

pytestmark = pytest.mark.skip


def test_that_creating_a_base_widget_works():
    widget = Labeller()  # noqa


def test_that_display_calls_display_func(mocker):
    mock_display = mocker.Mock()
    widget = Labeller(display_func=mock_display)
    widget._display("dummy feature")
    assert mock_display.call_count == 1


def test_that_slow_displays_trigger_render_processing(mocker):
    widget = Labeller(display_func=lambda _: time.sleep(0.6))
    mock_processing = mocker.patch.object(widget, "_render_processing")
    widget._display("dummy feature")
    assert mock_processing.call_count == 0
    widget._display("dummy feature")
    assert mock_processing.call_count == 1


def test_that_from_images_sets_correct_arguments():

    # default
    widget = Labeller.from_images()
    assert widget._display_func.func is display.image_display_func
    assert widget._display_func.keywords == {"imsize": "square"}

    # tuple
    widget = Labeller.from_images(image_size=(200, 100))
    assert widget._display_func.func is display.image_display_func
    assert widget._display_func.keywords == {"imsize": (200, 100)}

    # string
    widget = Labeller.from_images(
        features=np.array([[0, 0, 0, 0], [0, 0, 0, 0]]), image_size="square"
    )
    assert widget._display_func.func is display.image_display_func
    assert widget._display_func.keywords == {"imsize": "square"}

    # features only
    widget = Labeller.from_images(
        features=np.array([[0, 0, 0, 0], [0, 0, 0, 0]])
    )
    assert widget._display_func.func is display.image_display_func
    assert widget._display_func.keywords == {"imsize": "square"}

    # error
    with pytest.raises(ValueError):
        widget = Labeller.from_images(
            features=np.array([[0, 0, 0, 0, 5], [0, 0, 0, 0, 5]])
        )

    with pytest.raises(TypeError):
        widget = Labeller.from_images(
            features=[[0, 0, 0, 0, 5], [0, 0, 0, 0, 5]]
        )


def test_that_ipython_display_displays_layout(mocker):
    mock_display = mocker.patch("IPython.display.display")
    widget = Labeller()
    widget._ipython_display_()
    assert mock_display.call_args == ((widget.layout,),)



import traitlets

import pytest
from hypothesis import strategies as st
from hypothesis import given, settings

from superintendent.controls.buttongroup import ButtonGroup, ButtonWithHint


@settings(deadline=None)
@given(options=st.lists(st.text(), unique=True))
def test_that_options_are_set(options):

    widget = ButtonGroup(options)
    assert widget.options == options
    assert pytest.helpers.same_elements(widget.buttons.keys(), options)
    assert pytest.helpers.same_elements(
        [button.description for button in widget.buttons.values()], options
    )

    widget = ButtonGroup([])
    widget.options = options
    assert widget.options == options
    assert pytest.helpers.same_elements(widget.buttons.keys(), options)
    assert pytest.helpers.same_elements(
        [button.description for button in widget.buttons.values()], options
    )


def test_that_on_click_adds_callables_to_execution_list(mocker):

    mock_callable = mocker.Mock()
    widget = ButtonGroup(["a", "b"])

    widget.on_click(mock_callable)
    assert widget.submission_functions == [mock_callable]

    widget._handle_click("test")

    assert mock_callable.call_args == (("test",),)

    with pytest.raises(ValueError):
        widget.on_click("non-callable")


def test_that_button_width_is_set_correctly():
    widget = ButtonGroup(["a", "b"])
    assert widget.button_width == "50%"

    widget.button_width = 0.2
    assert widget.button_width == "20%"

    widget.button_width = 200
    assert widget.button_width == "200px"

    widget.button_width = "87%"
    assert widget.button_width == "87%"

    with pytest.raises(traitlets.TraitError):
        widget.button_width = {"test-set", "lol"}


def test_that_enter_exit_for_output_get_called(mocker):
    mock_enter = mocker.patch("ipywidgets.Output.__enter__")
    mock_exit = mocker.patch("ipywidgets.Output.__exit__")

    button = ButtonWithHint("Hi", "50%")

    with button:
        pass

    assert mock_enter.call_count == 1
    assert mock_exit.call_count == 1



import pytest
import numpy as np

from superintendent.distributed.multiclass_labeller import MultiClassLabeller
from superintendent.multiclass_labeller import preprocess_multioutput
import superintendent.base
import superintendent.distributed.queueing
import superintendent.controls


def test_that_multiclass_labeller_creates_base_widget():

    widget = MultiClassLabeller()

    assert isinstance(widget, superintendent.base.Labeller)


def test_that_multiclass_labeller_has_expected_attributes():

    widget = MultiClassLabeller(options=("a", "b"))

    assert isinstance(
        widget.queue, superintendent.distributed.queueing.DatabaseQueue
    )
    assert isinstance(
        widget.input_widget, superintendent.controls.MulticlassSubmitter
    )

    assert pytest.helpers.same_elements(
        widget.input_widget.options, ("a", "b")
    )


def test_preprocessing(mocker):

    test_x = mocker.Mock()
    test_y = [["a"], ["a", "b"], ["b"]]

    x, y = preprocess_multioutput(test_x, test_y)

    assert x is test_x
    assert isinstance(y, np.ndarray)
    assert y.shape == (3, 2)
    assert np.all(y == np.array([[1, 0], [1, 1], [0, 1]]))

    x, y = preprocess_multioutput(test_x, None)
    assert x is test_x
    assert y is None


def test_that_preprocessing_is_set_correctly(mocker):

    test_x = None
    test_y = [["a"], ["a", "b"], ["b"]]

    widget = MultiClassLabeller()
    assert widget.model_preprocess is preprocess_multioutput

    test_preprocessor = mocker.MagicMock()
    test_preprocessor.return_value = (test_x, test_y)

    widget = MultiClassLabeller(model_preprocess=test_preprocessor)
    # assert it's been wrapped
    assert widget.model_preprocess is not test_preprocessor

    x, y = widget.model_preprocess(test_x, test_y)

    # check that original preprocessor was called
    test_preprocessor.assert_called_once_with(test_x, test_y)

    assert isinstance(y, np.ndarray)
    assert y.shape == (3, 2)
    assert np.all(y == np.array([[1, 0], [1, 1], [0, 1]]))



import pathlib
import numpy as np
from PIL import Image
from superintendent.display import (
    default_display_function,
    image_display_function,
)

TEST_ARRAY = np.random.randint(0, 256, (20, 20, 3), dtype=np.uint8)
TEST_ARRAY[:5, :] = 0
TEST_ARRAY[-5:, :] = 255

TEST_IMG = Image.fromarray(TEST_ARRAY)


def test_default_display(mocker):
    disp_mocker = mocker.patch("IPython.display.display")
    default_display_function("test data")
    disp_mocker.assert_any_call("test data")


def test_image_function_opens_strings_and_paths(mocker):

    test_path_str = "./test-img.png"
    test_path_path = pathlib.Path(test_path_str)

    mock_open = mocker.patch("PIL.Image.open")
    mock_open.return_value = TEST_IMG

    disp_mocker = mocker.patch("IPython.display.display")

    image_display_function(test_path_str, fit_into=(20, 20))
    assert mock_open.call_args[0] == (test_path_path,)
    assert disp_mocker.call_count == 1
    assert isinstance(disp_mocker.call_args[0][0], Image.Image)

    image_display_function(test_path_path, fit_into=(20, 20))
    assert mock_open.call_args[0] == (test_path_path,)
    assert disp_mocker.call_count == 2
    assert isinstance(disp_mocker.call_args[0][0], Image.Image)


def test_image_function_converts_numpy_arrays(mocker):

    disp_mocker = mocker.patch("IPython.display.display")

    test_array = np.random.rand(20, 20)
    image_display_function(test_array, fit_into=(20, 20))

    disp_mocker.assert_called_once()

    img = disp_mocker.call_args[0][0]

    assert isinstance(img, Image.Image)
    assert img.size == (20, 20)


def test_image_function_resizes(mocker):

    disp_mocker = mocker.patch("IPython.display.display")

    image_display_function(TEST_IMG, fit_into=(40, 40))

    disp_mocker.assert_called_once()

    img = disp_mocker.call_args[0][0]

    assert isinstance(img, Image.Image)
    assert img.size == (40, 40)

    assert np.allclose(
        np.array(img.resize((20, 20))),
        np.array(TEST_IMG.resize((40, 40)).resize((20, 20))),
    )



import pathlib
from functools import singledispatch
from typing import Tuple

import IPython.display
import numpy as np
from PIL import Image, ImageOps


@singledispatch
def image_display_function(image, fit_into=(500, 500)) -> None:
    """Display an image.

    Parameters
    ----------
    image : Pillow.Image.Image, np.ndarray, str, pathlib.Path
        a Pillow / PIL image, or the array data for it, or a path to an image.
    fit_into : tuple, optional
        What size the image should fit into, by default (500, 500). The image
        is scaled so that no side is larger than the corresponding pixels given
        here.
    """
    raise NotImplementedError(
        "You passed an object of type {}, but image_display_function "
        "expects a path, URL or numpy array."
    )


@image_display_function.register
def _image_display_function_path(
    image: pathlib.Path, fit_into: Tuple[int, int] = (500, 500)
) -> None:
    """Path -> Image"""
    image = Image.open(image)
    image_display_function(image)


@image_display_function.register
def _image_display_function_str(
    image: str, fit_into: Tuple[int, int] = (500, 500)
) -> None:
    """str -> Path"""
    path_image: pathlib.Path = pathlib.Path(image)
    image_display_function(path_image, fit_into=fit_into)


@image_display_function.register
def _image_display_function_array(
    image: np.ndarray, fit_into: Tuple[int, int] = (500, 500)
) -> None:
    """np.ndarray -> Image"""
    image = 255 * image / image.max()
    image = image.astype(np.uint8)
    image = Image.fromarray(image)

    image_display_function(image, fit_into=fit_into)


@image_display_function.register
def _image_display_function_pillow(
    image: Image.Image, fit_into: Tuple[int, int] = (500, 500)
):
    """Image -> display & finish"""
    if image.size < fit_into:
        factor = max(s1 / s2 for s1, s2 in zip(image.size, fit_into))
        image = image.resize((int(s / factor) for s in image.size))

    image = ImageOps.autocontrast(image)

    IPython.display.display(image)



import pytest

pytestmark = pytest.mark.skip

# import itertools
# import numpy as np
# import pandas as pd
# import ipyevents

# from superintendent import ClusterSupervisor

# TEST_ARRAY = np.array([[1, 2, 3], [1, 2, 3], [1, 2, 3], [1, 2, 3]])
# TEST_LABELS = np.array([1, 1, 2, 2])
# TEST_REPRESENTATIVENESS = np.array([0.5, 0.4, 0.6, 0.9])


# def test_that_creating_a_widget_works():
#     widget = ClusterSupervisor(TEST_ARRAY, TEST_LABELS)  # noqa: F841


# def test_features_and_labels_are_passed_on(mocker):
#     mock_queue = mocker.patch(
#         "superintendent.clustersupervisor.ClusterLabellingQueue"
#     )
#     widget = ClusterSupervisor(TEST_ARRAY, TEST_LABELS)  # noqa: F841
#     mock_queue.assert_called_once_with(TEST_ARRAY, TEST_LABELS, None)


# def test_that_supplied_representativeness_is_passed_to_queue(mocker):
#     mock_queue = mocker.patch(
#         "superintendent.clustersupervisor.ClusterLabellingQueue"
#     )
#     widget = ClusterSupervisor(  # noqa: F841
#         TEST_ARRAY, TEST_LABELS, TEST_REPRESENTATIVENESS
#     )
#     mock_queue.assert_called_once_with(
#         TEST_ARRAY, TEST_LABELS, TEST_REPRESENTATIVENESS
#     )


# def test_that_display_is_called_with_array_data(mocker):
#     mock_display = mocker.Mock()
#     widget = ClusterSupervisor(
#         TEST_ARRAY,
#         TEST_LABELS,
#         TEST_REPRESENTATIVENESS,
#         display_func=mock_display,
#     )

#     assert pytest.helpers.exact_element_match(
#         mock_display.call_args[0][0], TEST_ARRAY[:2, :]
#     )

#     widget._annotation_loop.send({"value": "dummy", "source": "dummy"})

#     assert widget.new_clusters == {1: "dummy"}


# def test_that_sending_undo_into_iterator_calls_undo_on_queue(mocker):
#     mock_undo = mocker.patch(
#         "superintendent.clustersupervisor.ClusterLabellingQueue.undo"
#     )

#     widget = ClusterSupervisor(TEST_ARRAY, TEST_LABELS)
#     widget._annotation_loop.send({"source": "__undo__", "value": None})

#     assert mock_undo.call_count == 2


# def test_that_sending_skip_calls_no_queue_method(mocker):
#     mock_undo = mocker.patch(
#         "superintendent.clustersupervisor.ClusterLabellingQueue.undo"
#     )
#     mock_submit = mocker.patch(
#         "superintendent.clustersupervisor.ClusterLabellingQueue.submit"
#     )

#     widget = ClusterSupervisor(TEST_ARRAY, TEST_LABELS)
#     widget._annotation_loop.send({"source": "__skip__", "value": None})

#     assert mock_undo.call_count == 0
#     assert mock_submit.call_count == 0


# def test_that_progressbar_value_is_updated_and_render_finished_called(
#   mocker
# ):
#     mock_render_finished = mocker.patch(
#         "superintendent.clustersupervisor."
#         "ClusterSupervisor._render_finished"
#     )

#     widget = ClusterSupervisor(
#         TEST_ARRAY, TEST_LABELS, keyboard_shortcuts=True
#     )

#     assert widget.progressbar.value == 0
#     widget._annotation_loop.send({"source": "", "value": "dummy label"})
#     assert widget.progressbar.value == 0.5
#     widget._annotation_loop.send({"source": "", "value": "dummy label"})
#     assert widget.progressbar.value == 1
#     assert mock_render_finished.call_count == 1


# def test_that_the_event_manager_is_closed(mocker):

#     mock_event_manager_close = mocker.patch.object(ipyevents.Event, "close")

#     widget = ClusterSupervisor(
#         TEST_ARRAY, TEST_LABELS, keyboard_shortcuts=True
#     )
#     widget._annotation_loop.send({"source": "", "value": "dummy label"})
#     widget._annotation_loop.send({"source": "", "value": "dummy label"})

#     assert mock_event_manager_close.call_count == 1



import numpy as np
import pandas as pd


def _features_to_array(features: list):
    """Convert a list of features to a 2D array.

    Parameters
    ----------
    features : list
        A list of features to be converted to an array or dataframe.

    Returns
    -------
    features : Any
        The array of features.
    """

    if len(features) > 0:
        if all(isinstance(feature, pd.Series) for feature in features):
            features = pd.concat([item.to_frame().T for item in features])
        elif all(isinstance(feature, pd.DataFrame) for feature in features):
            features = pd.concat(features)
        elif all(isinstance(feature, np.ndarray) for feature in features):
            features = np.stack(features)

    return features



import pytest
from hypothesis import strategies as st
from hypothesis import given, settings

from superintendent.controls.hintedmultiselect import HintedMultiselect


@settings(deadline=None)
@given(options=st.lists(st.text(), unique=True))
def test_that_options_are_set(options):

    widget = HintedMultiselect(options)
    assert widget.options == options == list(widget.multi_select.options)

    widget = HintedMultiselect([])
    widget.options = options
    assert widget.options == options == list(widget.multi_select.options)


def test_that_values_are_read_correctly():

    widget = HintedMultiselect(["a", "b", "c"])
    assert widget.value == []

    widget.multi_select.value = ["a"]
    assert widget.value == ["a"]
    widget.multi_select.value = ["a", "b"]
    assert pytest.helpers.same_elements(widget.value, ["a", "b"])


def test_that_toggle_changes_values():

    widget = HintedMultiselect(["a", "b", "c"])
    assert widget.value == []

    widget._toggle("a")
    assert widget.value == ["a"]

    widget._toggle("a")
    assert widget.value == []


def test_that_reset_sets_all_values_to_false():

    widget = HintedMultiselect(["a", "b", "c"])
    assert widget.value == []

    widget._toggle("a")
    widget._toggle("b")
    assert widget.value == ["a", "b"]

    widget._reset()
    assert widget.value == []


def test_that_enter_exit_for_output_get_called(mocker):
    mock_enter = mocker.patch("ipywidgets.Output.__enter__")
    mock_exit = mocker.patch("ipywidgets.Output.__exit__")

    widget = HintedMultiselect(["a", "b", "c"])

    with widget.hints["a"]:
        pass

    assert mock_enter.call_count == 1
    assert mock_exit.call_count == 1



from functools import partial

import pytest

import hypothesis.extra.numpy as np_strategies
import numpy as np
import pandas as pd
from hypothesis import given
from hypothesis.extra.pandas import column, columns, data_frames, series
from hypothesis.strategies import (
    booleans,
    dictionaries,
    floats,
    integers,
    lists,
    one_of,
    recursive,
    text,
)
from superintendent.distributed.serialization import data_dumps, data_loads

guaranteed_dtypes = (
    np_strategies.boolean_dtypes()
    | np_strategies.integer_dtypes()
    | np_strategies.floating_dtypes()
    | np_strategies.unicode_string_dtypes()
)


json_array = recursive(
    floats(allow_nan=False) | integers() | text() | booleans(), lists
)
json_object = recursive(
    floats(allow_nan=False) | integers() | text() | booleans(),
    partial(dictionaries, text()),
)


def exact_element_match(a, b):
    if isinstance(a, np.ndarray) and isinstance(b, np.ndarray):
        try:
            return ((a == b) | (np.isnan(a) & np.isnan(b))).all()
        except TypeError:
            return (a == b).all()
    elif isinstance(a, pd.DataFrame) and isinstance(b, pd.DataFrame):
        a = a.reset_index(drop=True)
        b = b.reset_index(drop=True)
        return (
            ((a == b) | (a.isnull() & b.isnull())).all().all()
            or a.empty
            or b.empty
        )
    else:
        return all(
            [
                a_ == b_ or (np.isnan(a_) and np.isnan(b_))
                for a_, b_ in zip(a, b)
            ]
        )


@given(
    input_=np_strategies.arrays(
        guaranteed_dtypes, np_strategies.array_shapes()
    )
)
def test_numpy_array_serialisation(input_):
    serialised = data_dumps(input_)
    assert isinstance(serialised, str)
    deserialised = data_loads(serialised)
    assert isinstance(deserialised, type(input_))
    assert pytest.helpers.exact_element_match(input_, deserialised)


@given(
    input_=one_of(series(dtype=int), series(dtype=float), series(dtype=str))
)
def test_pandas_series_serialisation(input_):
    serialised = data_dumps(input_)
    assert isinstance(serialised, str)
    deserialised = data_loads(serialised)
    assert isinstance(deserialised, type(input_))
    assert pytest.helpers.exact_element_match(input_, deserialised)


@given(
    input_=one_of(
        data_frames(columns(3, dtype=int)),
        data_frames(columns(3, dtype=float)),
        data_frames(columns(3, dtype=str)),
        data_frames(
            [column(dtype=str), column(dtype=float), column(dtype=int)]
        ),
    )
)
def test_pandas_df_serialisation(input_):
    serialised = data_dumps(input_)
    assert isinstance(serialised, str)
    deserialised = data_loads(serialised)
    assert isinstance(deserialised, type(input_))
    assert pytest.helpers.exact_element_match(input_, deserialised)


@given(input_=one_of(json_array, json_object))
def test_normal_serialisation_dicts(input_):
    serialised = data_dumps(input_)
    assert isinstance(serialised, str)
    deserialised = data_loads(serialised)
    assert input_ == deserialised or np.isnan(input_)


def test_none_ok():
    assert data_dumps(None) is None
    assert data_loads(None) is None



import pytest

import ipywidgets

from hypothesis import strategies as st
from hypothesis import given, settings

from superintendent import controls
from superintendent.controls import togglebuttongroup
from superintendent.controls import hintedmultiselect


DUMMY_KEYUP_DICT_ENTER = {
    "altKey": False,
    "ctrlKey": False,
    "metaKey": False,
    "shiftKey": False,
    "type": "keyup",
    "timeStamp": 75212.30000001378,
    "code": "Enter",
    "key": "Enter",
    "location": 0,
    "repeat": False,
    "event": "keyup",
}

DUMMY_KEYUP_DICT_BACKSPACE = {
    "altKey": False,
    "ctrlKey": False,
    "metaKey": False,
    "shiftKey": False,
    "type": "keyup",
    "timeStamp": 75212.30000001378,
    "code": "Backspace",
    "key": "Backspace",
    "location": 0,
    "repeat": False,
    "event": "keyup",
}

DUMMY_KEYUP_DICT_ONE = {
    "altKey": False,
    "ctrlKey": False,
    "metaKey": False,
    "shiftKey": False,
    "type": "keyup",
    "timeStamp": 129242.40000001737,
    "code": "Digit1",
    "key": "1",
    "location": 0,
    "repeat": False,
    "event": "keyup",
}


@settings(deadline=None)
@given(options=st.lists(st.text()))
def test_that_options_are_set_correctly_when_instantiated(options):
    widget = controls.MulticlassSubmitter(options=options)
    assert widget.options == options
    assert widget.control_elements.options == widget.options


@settings(deadline=None)
@given(options=st.lists(st.text()))
def test_that_options_are_set_correctly_when_updated(options):
    widget = controls.MulticlassSubmitter()
    assert widget.options == []
    assert widget.control_elements.options == widget.options

    widget.options = options
    assert widget.options == options
    assert widget.control_elements.options == widget.options


@settings(deadline=None)
@given(options=st.lists(st.integers()))
def test_that_options_are_converted_to_strings(options):
    widget = controls.MulticlassSubmitter(options=options)
    assert widget.options == [str(option) for option in options]


def test_that_updating_options_triggers_compose(mocker):
    options = ["a", "b", "c"]
    widget = controls.MulticlassSubmitter()
    mock_compose = mocker.patch.object(widget, "_compose")
    widget.options = options
    assert mock_compose.called_once()


def test_that_on_submit_updates_submissions_functions():
    widget = controls.MulticlassSubmitter()
    widget.on_submit(print)
    assert widget.submission_functions == [print]


def test_that_on_submit_fails_when_not_given_callable():
    widget = controls.MulticlassSubmitter()
    with pytest.raises(ValueError):
        widget.on_submit("dummy")


def test_that_submit_passes_correct_values(mocker):
    widget = controls.MulticlassSubmitter(["a", "b"])
    mock_function = mocker.Mock()
    widget.on_submit(mock_function)

    widget._toggle_option("a")

    widget._submit(widget.submission_button)

    mock_function.assert_called_once()
    assert mock_function.call_args[0] == (["a"],)


def test_that_text_field_adds_option_and_toggles_it(mocker):
    widget = controls.MulticlassSubmitter(["a", "b"])

    assert "c" not in widget.options

    widget.other_widget.value = "c"
    widget._handle_new_option(widget.other_widget)

    assert "c" in widget.options
    assert widget.control_elements.value == ["c"]


def test_on_undo_args_get_called(mocker):
    widget = controls.MulticlassSubmitter(["a", "b"])
    mock_undo = mocker.MagicMock()
    widget.on_undo(mock_undo)
    widget._undo("dummy")

    mock_undo.assert_called_once()
    assert mock_undo.call_args[0] == tuple()


def test_on_skip_args_get_called(mocker):
    widget = controls.MulticlassSubmitter(["a", "b"])
    mock_skip = mocker.MagicMock()
    widget.on_skip(mock_skip)
    widget._skip("dummy")

    mock_skip.assert_called_once()
    assert mock_skip.call_args[0] == tuple()


def test_that_sort_options_sorts_options(mocker):
    options = ["d", "a", "b", "c"]
    widget = controls.MulticlassSubmitter(options)
    mock_compose = mocker.patch.object(widget, "_compose")
    widget._sort_options()

    assert widget.options == ["a", "b", "c", "d"]
    assert mock_compose.called_once()


def test_that_removing_core_options_fails():
    options = ["a", "b", "c"]
    widget = controls.MulticlassSubmitter(options)

    widget.remove_options(["a"])

    assert widget.options == options


def test_that_removing_added_options_works():
    options = ["a", "b", "c"]
    widget = controls.MulticlassSubmitter(options)

    widget.options = widget.options + ["d"]
    assert "d" in widget.options

    widget.remove_options(["d"])
    assert "d" not in widget.options


def test_that_a_button_group_is_created():
    options = ["a", "b", "c"]
    widget = controls.MulticlassSubmitter(options)
    assert isinstance(
        widget.control_elements, togglebuttongroup.ToggleButtonGroup
    )
    assert widget.options == widget.control_elements.options
    widget.options = ["a", "b", "c", "d"]
    assert widget.options == widget.control_elements.options


def test_that_a_multiselector_is_created():
    options = ["a", "b", "c", "d", "e", "f"]
    widget = controls.MulticlassSubmitter(options, max_buttons=3)
    assert isinstance(
        widget.control_elements, hintedmultiselect.HintedMultiselect
    )
    assert widget.options == widget.control_elements.options
    widget.options = ["a", "b", "c", "d"]
    assert widget.options == widget.control_elements.options


def test_that_a_textbox_is_created_or_not():
    options = ["a", "b", "c", "d", "e", "f"]
    widget = controls.MulticlassSubmitter(options)
    assert isinstance(widget.other_widget, ipywidgets.Text)

    widget = controls.MulticlassSubmitter(options, other_option=False)
    assert (
        isinstance(widget.other_widget, ipywidgets.HBox)
        and len(widget.other_widget.children) == 0
    )



import pytest  # noqa

import numpy as np
import pandas as pd

from hypothesis import given, settings
from hypothesis.strategies import (
    booleans,
    floats,
    integers,
    lists,
    tuples,
    one_of,
    sampled_from,
    text,
    composite,
)
from hypothesis.extra.pandas import data_frames, column, range_indexes
from hypothesis.extra.numpy import (
    arrays,
    scalar_dtypes,
    unsigned_integer_dtypes,
    datetime64_dtypes,
    floating_dtypes,
    integer_dtypes,
)

from hypothesis import HealthCheck

from superintendent.queueing.utils import _features_to_array


guaranteed_dtypes = one_of(
    scalar_dtypes(),
    unsigned_integer_dtypes(),
    datetime64_dtypes(),
    floating_dtypes(),
    integer_dtypes(),
)


@composite
def dataframe(draw):
    n_cols = draw(integers(min_value=1, max_value=20))
    dtypes = draw(
        lists(
            sampled_from([float, int, str]), min_size=n_cols, max_size=n_cols
        )
    )
    colnames = draw(
        lists(
            text() | integers(), min_size=n_cols, max_size=n_cols, unique=True
        )
    )
    return draw(
        data_frames(
            columns=[
                column(name=name, dtype=dtype)
                for dtype, name in zip(dtypes, colnames)
            ],
            index=range_indexes(min_size=1),
        )
    )


def exact_element_match(a, b):
    if isinstance(a, np.ndarray) and isinstance(b, np.ndarray):
        try:
            return ((a == b) | (np.isnan(a) & np.isnan(b))).all()
        except TypeError:
            return (a == b).all()
    elif isinstance(a, pd.DataFrame) and isinstance(b, pd.DataFrame):
        a = a.reset_index(drop=True)
        b = b.reset_index(drop=True)
        return (
            ((a == b) | (a.isnull() & b.isnull())).all().all()
            or a.empty
            or b.empty
        )
    else:
        return all(
            [
                a_ == b_ or (np.isnan(a_) and np.isnan(b_))
                for a_, b_ in zip(a, b)
            ]
        )


@given(inp=lists(floats() | integers() | text() | booleans()))
def test_list_round_trip(inp):
    assert exact_element_match(inp, _features_to_array(inp))


@given(
    inp=arrays(
        guaranteed_dtypes,
        tuples(
            integers(min_value=1, max_value=50),
            integers(min_value=1, max_value=50),
        ),
    )
)
@settings(suppress_health_check=(HealthCheck.too_slow,))
def test_array_round_trip(inp):
    inp_list = list(inp)
    assert exact_element_match(inp, _features_to_array(inp_list))


@given(inp=dataframe())
@settings(suppress_health_check=(HealthCheck.too_slow,))
def test_df_round_trip(inp):
    inp_list = [row for _, row in inp.iterrows()]
    if not inp.empty:
        assert exact_element_match(inp, _features_to_array(inp_list))


@given(inp=dataframe())
@settings(suppress_health_check=(HealthCheck.too_slow,))
def test_dfs_round_trip(inp):
    inp_list = [row.to_frame().T for _, row in inp.iterrows()]
    if not inp.empty:
        assert exact_element_match(inp, _features_to_array(inp_list))



"""Interactive machine learning supervision using databases to distribute."""

from .class_labeller import ClassLabeller
from .multiclass_labeller import MultiClassLabeller

__all__ = ["ClassLabeller", "MultiClassLabeller"]



"""Classes for arranging queues"""

from .base import BaseLabellingQueue
from .cluster_queue import ClusterLabellingQueue
from .in_memory import SimpleLabellingQueue

__all__ = [
    "SimpleLabellingQueue",
    "ClusterLabellingQueue",
    "BaseLabellingQueue",
]



import os
import os.path
from collections import Counter, OrderedDict
from contextlib import contextmanager

import numpy as np
import pandas as pd
import pytest
from hypothesis import HealthCheck, given, settings
import hypothesis.extra.numpy as np_strategies
from hypothesis.extra.pandas import column, data_frames
from hypothesis.strategies import (
    booleans,
    composite,
    floats,
    integers,
    lists,
    one_of,
    text,
    tuples,
)

from superintendent.distributed.queueing import DatabaseQueue

guaranteed_dtypes = one_of(
    np_strategies.scalar_dtypes(),
    np_strategies.unsigned_integer_dtypes(),
    np_strategies.datetime64_dtypes(),
    np_strategies.floating_dtypes(),
    np_strategies.integer_dtypes(),
)


@composite
def dataframe(draw):
    n_cols = draw(integers(min_value=1, max_value=20))
    dtypes = draw(
        lists(
            one_of(
                np_strategies.floating_dtypes(),
                np_strategies.integer_dtypes(),
                np_strategies.unicode_string_dtypes(),
            ),
            min_size=n_cols,
            max_size=n_cols,
        )
    )
    colnames = draw(
        lists(
            text() | integers(), min_size=n_cols, max_size=n_cols, unique=True
        )
    )
    return draw(
        data_frames(
            columns=[
                column(name=name, dtype=dtype)
                for dtype, name in zip(dtypes, colnames)
            ]
        )
    )


@contextmanager
def q_context():
    try:
        yield DatabaseQueue("sqlite:///testing.db")
    finally:
        if os.path.isfile("testing.db"):
            os.remove("testing.db")


@settings(deadline=None)
@given(input_=one_of(booleans(), floats(), integers(), text()))
def test_enqueueing_and_popping(input_):
    with q_context() as q:
        for i in range(1, 10):
            q.enqueue(input_)
            id_, datapoint = q.pop()
            assert id_ == i
            assert datapoint == input_ or (
                np.isnan(input_) and np.isnan(datapoint)
            )


@settings(deadline=None)
@given(inputs=lists(one_of(booleans(), floats(), integers(), text())))
def test_enqueue_many(inputs):
    n = len(inputs)
    with q_context() as q:
        q.enqueue_many(inputs)
        # assert we can pop everything:
        for _ in range(n):
            q.pop()
        # assert there's nothing else to pop:
        with pytest.raises(IndexError):
            q.pop()


@settings(suppress_health_check=(HealthCheck.too_slow,), deadline=None)
@given(inputs=dataframe())
def test_enqueue_dataframe(inputs):
    n = len(inputs)
    with q_context() as q:
        q.enqueue_many(inputs)
        # assert we can pop everything:
        for _ in range(n):
            id_, val = q.pop()
            assert isinstance(val, pd.Series)
        # assert there's nothing else to pop:
        with pytest.raises(IndexError):
            q.pop()
        # assert it re-constructs a df on list all
        if n > 0:
            ids, X, y = q.list_all()
            assert isinstance(X, pd.DataFrame) or len(X) == 0
            # assert it re-constructs a df on list uncomplete
            q.submit(ids[0], "hello")
            ids, X = q.list_uncompleted()
            assert isinstance(X, pd.DataFrame) or len(X) == 0
            # assert it re-constructs a df on list uncomplete
            ids, X, y = q.list_completed()
            assert isinstance(X, pd.DataFrame) or len(X) == 0


@settings(suppress_health_check=(HealthCheck.too_slow,), deadline=None)
@given(
    inputs=np_strategies.arrays(
        guaranteed_dtypes,
        tuples(
            integers(min_value=1, max_value=50),
            integers(min_value=1, max_value=50),
        ),
    )
)
def test_enqueue_array(inputs):
    n = inputs.shape[0]
    with q_context() as q:
        q.enqueue_many(inputs)
        # assert we can pop everything:
        for _ in range(n):
            id_, val = q.pop()
            assert isinstance(val, np.ndarray)
            assert len(val.shape) == 1
            assert val.size == inputs.shape[-1]
        # assert there's nothing else to pop:
        with pytest.raises(IndexError):
            q.pop()
        # assert it re-constructs a df on list all
        if n > 0:
            ids, X, y = q.list_all()
            assert isinstance(X, np.ndarray)


@settings(deadline=None)
@given(
    inputs=lists(one_of(booleans(), floats(), integers(), text())),
    labels=lists(text()),
)
def test_enqueue_with_labels(inputs, labels):

    if len(labels) > len(inputs):
        labels = labels[: len(inputs)]
    elif len(labels) < len(inputs):
        labels += [None] * (len(inputs) - len(labels))

    n = len(inputs) - (len(labels) - Counter(labels)[None])

    with q_context() as q:
        q.enqueue_many(inputs, labels)

        # assert we can pop everything where the label was not none:
        for _ in range(n):
            q.pop()
        # assert there's nothing else to pop:
        with pytest.raises(IndexError):
            q.pop()


@pytest.mark.skip
@settings(deadline=None)
@given(inputs=lists(one_of(booleans(), floats(), integers(), text())))
def test_enqueue_at_creation(inputs):
    n = len(inputs)
    with q_context() as q:
        assert len(q.data) == n
        # assert we can pop everything:
        for _ in range(n):
            q.pop()
        # assert there's nothing else to pop:
        with pytest.raises(IndexError):
            q.pop()


@settings(deadline=None)
@given(label1=text(), label2=text())
def test_submitting_text(label1, label2):
    with q_context() as q:
        q.enqueue(1)
        q.enqueue(2)
        with pytest.raises(ValueError):
            q.submit(1, label1)
        id_, val = q.pop()
        q.submit(id_, label1)
        assert q.progress == 0.5
        id_, val = q.pop()
        q.submit(id_, label2)
        assert q.progress == 1
        assert q.list_labels() == {label1, label2}


@settings(deadline=None)
@given(label1=text(), label2=text())
def test_submitting_list(label1, label2):
    with q_context() as q:
        q.enqueue(1)
        with pytest.raises(ValueError):
            q.submit(1, label1)
        id_, val = q.pop()
        q.submit(id_, [label1, label2])
        assert q.list_labels() == {label1, label2}


def test_reordering():
    inp = ["b", "a", "d", "c"]
    with q_context() as q:
        q.enqueue_many(inp)
        q.reorder(OrderedDict([(1, 1), (2, 0), (3, 3), (4, 2)]))

        id_, val = q.pop()
        assert val == "a" and id_ == 2
        id_, val = q.pop()
        assert val == "b" and id_ == 1
        id_, val = q.pop()
        assert val == "c" and id_ == 4
        id_, val = q.pop()
        assert val == "d" and id_ == 3


def test_iterating_over_queue():
    inps = [str(i) for i in range(50)]
    with q_context() as q:
        q.enqueue_many(inps)

        for i, (id_, val) in enumerate(q):
            assert i + 1 == id_


def test_length_of_queue():
    inps = [str(i) for i in range(50)]
    with q_context() as q:
        assert len(q) == 0
        q.enqueue_many(inps)
        assert len(q) == len(inps)


def test_progress():
    inps = [str(i) for i in range(50)]
    with q_context() as q:
        # assert my little hack for dividing by zero
        # assert np.isnan(q.progress)
        q.enqueue_many(inps)

        for i, (id_, val) in enumerate(q):
            assert q.progress == i / len(inps)
            q.submit(id_, str(i))
    with q_context() as q:
        assert np.isnan(q.progress)


@pytest.mark.skip
def test_shuffling():
    inps = [str(i) for i in range(50)]
    with q_context() as q:
        q.enqueue_many(inps)
        q.shuffle()
        # assert the order is not the same:
        assert not all([val == inp for inp, (id_, val) in zip(inps, q)])


def test_undo():
    inp = "input 1"
    with q_context() as q:
        q.enqueue(inp)
        id_, val = q.pop()
        q.submit(id_, "label 1")
        # ensure the queue is empty now:
        with pytest.raises(IndexError):
            q.pop()
        q.undo()
        # see if it's possible to pop now:
        id_, val = q.pop()
        assert val == "input 1"


def no_shared_members(a, b):
    return (set(a) & set(b)) == set()


@given(
    inputs=lists(one_of(booleans(), floats(), integers(), text()), min_size=5),
    labels=lists(text(), min_size=5),
)
def test_list_completed(inputs, labels):
    with q_context() as q:
        q.enqueue_many(inputs)

        popped_ids = []
        for i in range(5):
            id_, val = q.pop()
            q.submit(id_, labels[i])

            popped_ids.append(id_)

        ids, x, y = q.list_completed()

        assert len(ids) == 5
        # test that the popped IDs and completed IDs have the same members
        assert pytest.helpers.same_elements(ids, popped_ids)
        assert pytest.helpers.same_elements(y, labels[:5])


@given(
    inputs=lists(one_of(booleans(), floats(), integers(), text()), min_size=5),
    labels=lists(text(), min_size=5),
)
def test_list_uncompleted(inputs, labels):
    with q_context() as q:
        q.enqueue_many(inputs)

        popped_ids = []
        for i in range(5):
            id_, val = q.pop()
            q.submit(id_, labels[i])

            popped_ids.append(id_)

        ids, x = q.list_uncompleted()

        assert len(ids) == (len(inputs) - 5)
        assert q._unlabelled_count() == (len(inputs) - 5)
        # test that the popped IDs and completed IDs don't share members
        assert pytest.helpers.no_shared_members(ids, popped_ids)
        # assert pytest.helpers.same_elements(x, [inputs[idx] for idx in id])


@given(
    inputs=lists(one_of(booleans(), floats(), integers(), text()), min_size=5),
    labels=lists(text(), min_size=5),
)
def test_list_all(inputs, labels):
    with q_context() as q:
        q.enqueue_many(inputs)

        popped_ids = []
        for i in range(5):
            id_, val = q.pop()
            q.submit(id_, labels[i])

            popped_ids.append(id_)

        ids, x, y = q.list_all()

        assert len(ids) == len(inputs)
        assert all([label in labels for label in y if label is not None])
        assert all(
            [label is None or id_ in popped_ids for id_, label in zip(ids, y)]
        )
        assert Counter(y)[None] == (len(inputs) - 5)
        assert pytest.helpers.same_elements(ids, range(1, 1 + len(inputs)))
